from typing import Iterator, Iterable, List


def html_table(dictset: Iterator[dict], limit: int = 5):
    """
    Render the dictset as a HTML table.

    NOTE:
        This exhausts generators so is only recommended to be used on lists.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to render
        limit: integer (optional)
            The maximum number of record to show in the table, defaults to 5

    Returns:
        string (HTML table)
    """

    def _to_html_table(data, columns):

        yield '<table class="table table-sm">'
        for counter, record in enumerate(data):
            if counter == 0:
                yield '<thead class="thead-light"><tr>'
                for column in columns:
                    yield "<th>" + column + "<th>\n"
                yield "</tr></thead><tbody>"

            if (counter % 2) == 0:
                yield '<tr style="background-color:#F4F4F4">'
            else:
                yield "<tr>"
            for column in columns:
                yield "<td>" + str(record.get(column)) + "<td>\n"
            yield "</tr>"

        yield "</tbody></table>"

    rows = []
    columns = []  # type:ignore
    for i, row in enumerate(dictset):
        rows.append(row)
        columns = columns + list(row.keys())
        if (i + 1) == limit:
            break
    columns = set(columns)  # type:ignore

    import types

    footer = ""
    if isinstance(dictset, types.GeneratorType):
        footer = f"\n<p>top {limit} rows x {len(columns)} columns</p>"
        footer += "\nNOTE: the displayed records have been spent"
    if isinstance(dictset, list):
        footer = f"\n<p>{len(dictset)} rows x {len(columns)} columns</p>"

    return "".join(_to_html_table(rows, columns)) + footer


def ascii_table(dictset: Iterator[dict], limit: int = 5):
    """
    Render the dictset as a ASCII table.

    NOTE:
        This exhausts generators so is only recommended to be used on lists.

    Parameters:
        dictset: iterable of dictionaries
            The dictset to render
        limit: integer (optional)
            The maximum number of record to show in the table, defaults to 5

    Returns:
        string (ASCII table)
    """
    result = []
    columns: dict = {}
    cache = []

    # inspect values
    for count, row in enumerate(dictset):
        if count == limit:
            break

        cache.append(row)
        for k, v in row.items():
            length = max(len(str(v)), len(k))
            if length > columns.get(k, 0):
                columns[k] = length

    # draw table
    bars = []
    for header, width in columns.items():
        bars.append("─" * (width + 2))

    # display headers
    result.append("┌" + "┬".join(bars) + "┐")
    result.append("│" + "│".join([k.center(v + 2) for k, v in columns.items()]) + "│")
    result.append("├" + "┼".join(bars) + "┤")

    # display values
    for row in cache:
        result.append(
            "│"
            + "│".join([str(v).center(columns[k] + 2) for k, v in row.items()])
            + "│"
        )

    # display footer
    result.append("└" + "┴".join(bars) + "┘")

    return "\n".join(result)


BAR_CHARS = [r" ", r"▁", r"▂", r"▃", r"▄", r"▅", r"▆", r"▇", r"█"]


def draw_histogram_bins(bins: List[int]):
    """
    Draws a pre-binned set off histogram data
    """
    mx = max(bins)
    bar_height = mx / 8
    if bar_height == 0:
        return " " * len(bins)

    histogram = ""
    for value in bins:
        if value == 0:
            histogram += BAR_CHARS[0]
        else:
            height = int(value / bar_height)
            histogram += BAR_CHARS[height]

    return histogram


def histogram(values: Iterable[int], number_of_bins: int = 10):
    """ """
    bins = [0] * number_of_bins
    mn = min(values)
    mx = max(values)
    interval = ((1 + mx) - mn) / (number_of_bins - 1)

    for v in values:
        bins[int(v / interval)] += 1
    return draw_histogram_bins(bins)



"""
This is an initial implementation of a filtering mechanism to be applied
when reading data. There is an existing function-based implemented
however, as this makes it difficult to understand the fields being
used to filter on, when data is indexed it won't be able to take 
advantage of the filters.

This current implementation is a step toward that, it does not use
any indices at this point.
"""
import re
import operator
from functools import lru_cache
from typing import Optional, Iterable, List, Tuple, Union
from ....errors import InvalidSyntaxError
from ....logging import get_logger


# https://codereview.stackexchange.com/a/248421
_special_regex_chars = {ch: "\\" + ch for ch in ".^$*+?{}[]|()\\"}


@lru_cache(4)
def _sql_like_fragment_to_regex(fragment):
    """
    Allows us to accepts LIKE statements to search data
    """
    # https://codereview.stackexchange.com/a/36864/229677
    safe_fragment = "".join([_special_regex_chars.get(ch, ch) for ch in fragment])
    return re.compile("^" + safe_fragment.replace("%", ".*?").replace("_", ".") + "$")


# set of operators we can interpret not in the `operator` module
def _like(x, y):
    return _sql_like_fragment_to_regex(y.lower()).match(str(x).lower())


def _in(x, y):
    return x in y


def _nin(x, y):
    return x not in y


def _con(x, y):
    return y in x


def _ncon(x, y):
    return y not in x


def true(x):
    return True


# convert text representation of operators to functions
OPERATORS = {
    "=": operator.eq,
    "==": operator.eq,
    "is": operator.eq,
    "!=": operator.ne,
    "<": operator.lt,
    ">": operator.gt,
    "<=": operator.le,
    ">=": operator.ge,
    "like": _like,
    "in": _in,
    "!in": _nin,
    "not in": _nin,
    "contains": _con,
    "!contains": _ncon,
}


def evaluate(predicate: Union[tuple, list], record: dict) -> bool:
    """
    This is the evaluation routine for the Filter class.

    Implements a DNF (Disjunctive Normal Form) interpretter. Predicates in the
    same list are joined with an AND (*all()*) and predicates in adjacent lists
    are joined with an OR (*any()*). This allows for non-trivial filters to be
    written with just _tuples_ and _lists_.

    The predicates are in _tuples_ in the form (`key`, `op`, `value`) where the
    `key` is the value looked up from the record, the `op` is the operator and
    the `value` is a literal.
    """
    # No filter doesn't filter
    if predicate is None:  # pragma: no cover
        return True

    # If we have a tuple extract out the key, operator and value
    # and do the evaluation
    if isinstance(predicate, tuple):
        key, op, value = predicate
        if key in record:
            return OPERATORS[op](record[key], value)
        return False

    if isinstance(predicate, list):
        # Are all of the entries tuples?
        # We AND them together (_all_ are True)
        if all([isinstance(p, tuple) for p in predicate]):
            return all([evaluate(p, record) for p in predicate])

        # Are all of the entries lists?
        # We OR them together (_any_ are True)
        if all([isinstance(p, list) for p in predicate]):
            return any([evaluate(p, record) for p in predicate])

        # if we're here the structure of the filter is wrong
        raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover

    raise InvalidSyntaxError("Unable to evaluate Filter")  # pragma: no cover


def get_indexable_filter_columns(predicate):
    """
    Returns all of the columns in a filter which the operation benefits
    from an index

    This creates an list of tuples of (field,value) that we can feed to the
    index search.
    """
    INDEXABLE_OPS = {"=", "==", "is", "in", "contains"}
    if predicate is None:
        return []
    if isinstance(predicate, tuple):
        key, op, value = predicate
        if op in INDEXABLE_OPS:
            return [
                (
                    key,
                    value,
                )
            ]
    if isinstance(predicate, list):
        if all([isinstance(p, tuple) for p in predicate]):
            return [
                (
                    k,
                    v,
                )
                for k, o, v in predicate
                if o in INDEXABLE_OPS
            ]
        if all([isinstance(p, list) for p in predicate]):
            columns = []
            for p in predicate:
                columns += get_indexable_filter_columns(p)
            return columns
    return []  # pragma: no cover


class Filters:

    __slots__ = ("empty_filter", "predicates")

    def __init__(self, filters: Optional[List[Tuple[str, str, object]]] = None):
        """
        A class to supporting filtering data.

        Parameters:
            filters: list of tuples
                Each tuple has format: (`key`, `op`, `value`). When run the
                filter will extract the `key` field from the dictionary and
                compare to the `value` using the operator `op`. Multiple
                filters are treated as AND, lists of ANDs are treated as ORs.
                The supported `op` values are: `=` or `==`, `!=`, `<`, `>`,
                `<=`, `>=`, `in`, `!in` (not in), `contains`, `!contains`
                (does not contain) and `like`. If the `op` is `in` or `!in`,
                the `value` must be a collection such as a _list_, a _set_
                or a _tuple_.
                `like` performs similar to the SQL operator, `%` is a
                multi-character wildcard and `_` is a single character
                wildcard.

        Examples:
            filters = Filters([('name', '==', 'john')])
            filters = Filters([('name', '!=', 'john'),('name', '!=', 'tom')])

        """
        if filters:
            self.predicates = filters
            self.empty_filter = False
            # record the filters, will help optimize indicies later
            get_logger().info({"filter_columns": self._get_filter_columns(filters)})
        else:
            self.empty_filter = True

    def _get_filter_columns(self, predicate):
        if predicate is None:
            return []
        if isinstance(predicate, tuple):
            key, op, value = predicate
            return [key]
        if isinstance(predicate, list):
            if all([isinstance(p, tuple) for p in predicate]):
                return [k for k, o, v in predicate]
            if all([isinstance(p, list) for p in predicate]):
                columns = []
                for p in predicate:
                    columns += self._get_filter_columns(p)
                return columns
        return []  # pragma: no cover

    def filter_dictset(self, dictset: Iterable[dict]) -> Iterable:
        """
        Tests each entry in a Iterable against the filters

        Parameters:
            dictset: iterable of dictionaries
                The dictset to process

        Yields:
            dictionary
        """
        if self.empty_filter:
            yield from dictset
        else:
            for record in dictset:
                if evaluate(self.predicates, record):
                    yield record



"""
MinIO Bin Writer

May support AWS S3 - untested
"""
import time
import io
from .base_bin import BaseBin
from ...errors import MissingDependencyError

try:
    from minio import Minio  # type:ignore

    minio_installed = True
except ImportError:  # pragma: no cover
    minio_installed = False


class MinioBin(BaseBin):
    def __init__(
        self,
        bin_name: str,
        end_point: str,
        bucket: str,
        path: str,
        access_key: str,
        secret_key: str,
        secure: bool = True,
    ):

        if not minio_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`minio` is missing, please install or include in requirements.txt"
            )

        self.client = Minio(end_point, access_key, secret_key, secure=secure)
        self.bucket = bucket
        self.path = path
        self.name = bin_name

    def __str__(self) -> str:
        return self.name

    def __call__(self, record: str, id_: str = ""):

        filename = f"{self.path}/{self._date_part()}/{id_}{time.time_ns()}.txt"
        record_bytes = record.encode("utf-8")
        record_stream = io.BytesIO(record_bytes)

        self.client.put_object(
            self.bucket, filename, record_stream, length=len(record_bytes)
        )

        return filename



"""
MinIo Reader - also works with AWS
"""
import io
from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...utils import paths, common
from ...errors import MissingDependencyError

try:
    from minio import Minio  # type:ignore

    minio_installed = True
except ImportError:  # pragma: no cover
    minio_installed = False


class MinIoReader(BaseInnerReader):

    RULES = [
        {"name": "end_point", "required": True},
        {"name": "access_key", "required": True},
        {"name": "secret_key", "required": True},
        {"name": "secure", "required": True},
    ]

    def __init__(self, end_point: str, access_key: str, secret_key: str, **kwargs):

        if not minio_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`minio` is missing, please install or include in requirements.txt"
            )

        super().__init__(**kwargs)
        secure = kwargs.get("secure", True)
        self.minio = Minio(end_point, access_key, secret_key, secure=secure)

    def get_blobs_at_path(self, path):
        bucket, object_path, _, _ = paths.get_parts(path)
        for cycle_date in common.date_range(self.start_date, self.end_date):
            cycle_path = paths.build_path(path=object_path, date=cycle_date)
            blobs = self.minio.list_objects(
                bucket_name=bucket, prefix=cycle_path, recursive=True
            )

            yield from [
                bucket + "/" + blob.object_name
                for blob in blobs
                if not blob.object_name.endswith("/")
            ]

    def get_blob_stream(self, blob_name: str) -> io.IOBase:
        bucket, object_path, name, extension = paths.get_parts(blob_name)
        stream = self.minio.get_object(bucket, object_path + name + extension).read()

        io_stream = io.BytesIO(stream)
        return io_stream



"""
A Flow is a simplified Graph, this originally used NetworkX but it was replaced
with a bespoke graph implementation as the NetworkX implementation was being
monkey-patches to make it easier to use. The decision was made to write a
specialized, albeit simple, graph library that didn't require monkey-patching.
"""
from typing import List
from .bins import FileBin, GoogleCloudStorageBin, MinioBin
from .flow_runner import FlowRunner
from .internals.base_operator import BaseOperator
from ..errors import FlowError


class Flow:
    def __init__(self):
        """
        Flow represents Directed Acyclic Graphs which are used to describe data
        pipelines.
        """
        self.nodes = {}
        self.edges = []
        self.has_run = False

    def add_operator(self, name, operator):
        """
        Add a step to the DAG

        Parameters:
            name: string
                The name of the step, must be unique
            Operator: BaseOperator
                The Operator
        """
        self.nodes[name] = operator

    def link_operators(self, source_operator, target_operator):
        """
        Link steps in a flow.

        Parameters:
            source_operator: string
                The name of the source step
            target_operator: string
                The name of the target step
        """
        edge = (source_operator, target_operator)
        if edge not in self.edges:
            self.edges.append((source_operator, target_operator))

    def get_outgoing_links(self, name):
        """
        Get the names of outgoing links from a given step.

        Paramters:
            name: string
                The name of the step to search from
        """
        retval = {target for source, target in self.edges if source == name}
        return sorted(retval)

    def get_exit_points(self):
        """
        Get steps in the flow with no outgoing steps.
        """
        sources = {source for source, target in self.edges}
        retval = {target for source, target in self.edges if target not in sources}
        return sorted(retval)

    def get_entry_points(self):
        """
        Get steps in the flow with no incoming steps.
        """
        targets = {target for source, target in self.edges}
        retval = {source for source, target in self.edges if source not in targets}
        return sorted(retval)

    def is_acyclic(self):
        # cycle over the graph removing a layer of exits each cycle
        # if we have nodes but no exists, we're cyclic
        my_edges = self.edges.copy()

        while len(my_edges) > 0:
            # find all of the exits
            sources = {source for source, target in my_edges}
            exits = {target for source, target in my_edges if target not in sources}

            if len(exits) == 0:
                return False

            # remove the exits
            new_edges = [
                (source, target) for source, target in my_edges if target not in exits
            ]
            my_edges = new_edges
        return True

    def get_operator(self, name):
        """
        Get the Operator class by name.

        Parameters:
            name: string
                The name of the step
        """
        return self.nodes.get(name)

    def merge(self, assimilatee):
        """
        Merge a flow into the current flow.

        Parameters:
            assimilatee: Flow
                The flow to assimilate into the current flows
        """
        self.nodes = {**self.nodes, **assimilatee.nodes}
        self.edges += assimilatee.edges
        self.edges = list(set(self.edges))

    def attach_writers(self, writers: List[dict]):

        for writer in writers:
            name = writer.get("name")
            class_name = writer.get("class")

            if class_name == "gcs":
                writer = GoogleCloudStorageBin(  # type: ignore
                    bin_name=name,  # type: ignore
                    project=writer.get("project"),  # type: ignore
                    bucket=writer.get("bucket"),  # type: ignore
                    path=writer.get("path"),  # type: ignore
                )
                self._attach_writer(writer)

            if class_name == "file" or class_name == "disk":
                writer = FileBin(  # type: ignore
                    bin_name=name, path=writer.get("path")  # type: ignore
                )  # type: ignore
                self._attach_writer(writer)

            if class_name == "minio":
                writer = MinioBin(  # type: ignore
                    bin_name=name,  # type: ignore
                    end_point=writer.get("end_point"),  # type: ignore
                    bucket=writer.get("bucket"),  # type: ignore
                    path=writer.get("path"),  # type: ignore
                    access_key=writer.get("access_key"),  # type: ignore
                    secret_key=writer.get("secret_key"),  # type: ignore
                    secure=writer.get("secure", True),
                )  # type: ignore
                self._attach_writer(writer)

    def _attach_writer(self, writer):
        """
        Attach the writer to each node in the flow
        """
        from ..logging import get_logger

        try:
            for operator_name in self.nodes:
                operator = self.get_operator(operator_name)
                setattr(operator, str(writer.name), writer)
            return True
        except Exception as err:
            get_logger().error(
                f"Failed to add writer to flow - {type(err).__name__} - {err}"
            )
            return False

    def _validate_flow(self):
        from ..operators import EndOperator

        # flow must be more than one item long
        if len(self.nodes) <= 1:
            raise FlowError(
                "Flow failed validation - Flows must have more than one Operator"
            )

        # flow paths must end with end operators
        if not all(
            [
                isinstance(self.get_operator(node), EndOperator)
                for node in self.get_exit_points()
            ]
        ):
            raise FlowError(
                "Flow failed validation - Flows must end with an EndOperator"
            )

        # flows must be acyclic
        if not self.is_acyclic():
            raise FlowError("Flow failed validation - Flows must be acyclic")

    def __enter__(self):
        if self.has_run:
            raise FlowError(
                "Flows can only have a single runner, either loop after creating the runner or build the flow again."
            )
        self._validate_flow()
        return FlowRunner(self)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """
        Finalize concludes the flow and returns the sensor information
        """
        from ..logging import get_logger

        # determine if we're closing because we had an error condition
        context = {}
        has_failure = False
        if exc_type:
            has_failure = exc_type.__name__ == "SystemExit"
        context["mabel:errored"] = has_failure

        FlowRunner(self)(BaseOperator.sigterm(), context)
        for operator_name in self.nodes.keys():
            operator = self.get_operator(operator_name)
            if operator:
                get_logger().audit(operator.read_sensors())
        self.has_run = True

    def __repr__(self):
        if not self.is_acyclic():
            return "Flow: cannot represent cyclic flows"
        return "\n".join(list(self._draw()))

    def __str__(self) -> str:
        return self.get_entry_points().pop()

    def _draw(self):
        for entry in self.get_entry_points():
            yield (f"{str(entry)}")
            t = self._tree(entry, "")
            yield ("\n".join(t))

    def _tree(self, node, prefix=""):

        space = "    "
        branch = " │  "
        tee = " ├─ "
        last = " └─ "

        contents = self.get_outgoing_links(node)
        # contents each get pointers that are ├── with a final └── :
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, child_node in zip(pointers, contents):
            yield prefix + pointer + str(child_node)
            if len(self.get_outgoing_links(node)) > 0:
                # extend the prefix and recurse:
                extension = branch if pointer == tee else space
                # i.e. space because last, └── , above so no more |
                yield from self._tree(str(child_node), prefix=prefix + extension)



import functools
import inspect
from ..logging import get_logger
from ..errors import InvalidReaderConfigError


def get_levenshtein_distance(word1, word2):
    """
    https://en.wikipedia.org/wiki/Levenshtein_distance
    :param word1:
    :param word2:
    :return:
    """
    word2 = word2.lower()
    word1 = word1.lower()
    matrix = [[0 for x in range(len(word2) + 1)] for x in range(len(word1) + 1)]

    for x in range(len(word1) + 1):
        matrix[x][0] = x
    for y in range(len(word2) + 1):
        matrix[0][y] = y

    for x in range(1, len(word1) + 1):
        for y in range(1, len(word2) + 1):
            if word1[x - 1] == word2[y - 1]:
                matrix[x][y] = min(
                    matrix[x - 1][y] + 1, matrix[x - 1][y - 1], matrix[x][y - 1] + 1
                )
            else:
                matrix[x][y] = min(
                    matrix[x - 1][y] + 1, matrix[x - 1][y - 1] + 1, matrix[x][y - 1] + 1
                )

    return matrix[len(word1)][len(word2)]


class validate:
    def __init__(self, rules):
        self.rules = rules

    def add_rule(self, rule):
        self.rules.append(rule)

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):

            has_errors = False

            my_rules = self.rules.copy()

            # get the names for all of the parameters
            func_parameters = list(inspect.signature(func).parameters)
            entered_parameters = kwargs.copy()
            for index, parameter in enumerate(args):
                entered_parameters[func_parameters[index]] = parameter

            # add rules from parameters
            for new_rules in [
                v.RULES for k, v in entered_parameters.items() if hasattr(v, "RULES")
            ]:
                my_rules += new_rules

            # check for spelling mistakes
            valid_parameters = {item["name"] for item in my_rules}
            for not_on_list in [
                item for item in entered_parameters if item not in valid_parameters
            ]:
                suggestion = []
                for valid in valid_parameters:
                    if get_levenshtein_distance(not_on_list, valid) <= 2:
                        suggestion.append(valid)
                if len(suggestion):
                    get_logger().error(
                        {
                            "error": "unknown parameter",
                            "function": func.__qualname__,
                            "parameter": not_on_list,
                            "closest valid option(s)": suggestion,
                        }
                    )
                else:
                    get_logger().error(
                        {
                            "error": "unknown parameter",
                            "function": func.__qualname__,
                            "parameter": not_on_list,
                        }
                    )
                has_errors = True

            # check for missing required paramters
            required_parameters = {
                item.get("name") for item in my_rules if item.get("required")
            }
            missing_required_parameters = [
                param
                for param in required_parameters
                if param not in entered_parameters
            ]
            if len(missing_required_parameters):
                get_logger().error(
                    {
                        "error": "missing required parameters",
                        "function": func.__qualname__,
                        "missing_paramters": missing_required_parameters,
                    }
                )
                has_errors = True

            # warnings
            warninged_parameters = {
                item.get("name"): item.get("warning")
                for item in my_rules
                if item.get("warning")
            }
            for parameter, warning in warninged_parameters.items():
                if parameter in entered_parameters:
                    get_logger().warning(warning)

            # toxic
            for param in [rule for rule in self.rules if rule.get("incompatible_with")]:
                if (
                    param.get("name") in entered_parameters
                    and entered_parameters[param.get("name")]
                ):
                    toxic = [
                        t
                        for t in param["incompatible_with"]
                        if t in entered_parameters.keys()
                    ]
                    if toxic:
                        has_errors = True
                        get_logger().error(
                            {
                                "error": "invalid combination of parameters",
                                "function": func.__qualname__,
                                "parameter": param.get("name", ""),
                                "incompatible with": toxic,
                            }
                        )

            if has_errors:
                raise InvalidReaderConfigError("Reader has invalid parameters")

            return func(*args, **kwargs)

        return wrapper



# google-cloud-logging
# pydantic
import os
import logging
import datetime
from typing import Union, Optional
from .levels import LEVELS, LEVELS_TO_STRING
from ..utils import is_running_from_ipython, safe_field_name

try:
    from google.cloud import logging as stackdriver  # type:ignore
    from google.cloud.logging import DESCENDING  # type:ignore
except ImportError:
    stackdriver = None  # type:ignore


LOG_SINK = "MABEL"


def extract_caller():
    import traceback
    import os.path

    frames = traceback.extract_stack()
    if len(frames) < 4:
        return "<unknown>", "<unknown>", -1
    frame = frames[len(frames) - 4]
    head, tail = os.path.split(frame.filename)
    return frame.name, tail, frame.lineno


class GoogleLogger(object):
    @staticmethod
    def supported():
        if not stackdriver:
            return False
        if os.environ.get("IGNORE_STACKDRIVER", False):
            return False
        if is_running_from_ipython():
            return False
        global LOG_SINK
        LOG_SINK = os.environ.get("LOG_SINK", LOG_SINK)
        return True

    @staticmethod
    def write_event(
        message: Union[str, dict],
        system: Optional[str] = None,
        severity: Optional[int] = logging.DEBUG,
    ):

        from .create_logger import LOG_NAME

        labels = {}
        if system:
            labels["system"] = system
        method, module, line = extract_caller()
        labels["method"] = method
        labels["module"] = module
        labels["line"] = str(line)

        if os.environ.get("DUAL_LOG", False) or os.environ.get("IGNORE_STACKDRIVER"):
            print(
                f"{LOG_NAME} | {LEVELS_TO_STRING.get(severity, 'UNKNOWN')} | {datetime.datetime.now().isoformat()} | {method}() | {module}:{line} | {message}"  # type:ignore
            )

        if os.environ.get("IGNORE_STACKDRIVER"):
            return

        client = stackdriver.Client()
        logger = client.logger(safe_field_name(LOG_SINK))

        if isinstance(message, dict):
            logger.log_struct(
                info=message,
                severity=LEVELS_TO_STRING.get(severity),  # type:ignore
                labels=labels,
            )
        else:
            logger.log_text(
                text=f"{message}",
                severity=LEVELS_TO_STRING.get(severity),  # type:ignore
                labels=labels,
            )

    def __init__(self):
        # rewrite one of the levels
        LEVELS_TO_STRING[LEVELS.AUDIT] = "NOTICE"

        self.level = int(os.environ.get("LOGGING_LEVEL", 25))
        self.debug = self.create_logger(LEVELS.DEBUG)
        self.info = self.create_logger(LEVELS.INFO)
        self.warning = self.create_logger(LEVELS.WARNING)
        self.error = self.create_logger(LEVELS.ERROR)
        self.audit = self.create_logger(LEVELS.AUDIT)
        self.alert = self.create_logger(LEVELS.ALERT)

    def create_logger(self, level):
        from .create_logger import LOG_NAME

        def base_logger(message):
            GoogleLogger.write_event(message=message, system=LOG_NAME, severity=level)

        def do_nothing(message):
            pass

        if level > self.level:
            return base_logger
        else:
            return do_nothing

    def setLevel(self, level):
        self.level = level
        self.debug = self.create_logger(LEVELS.DEBUG)
        self.info = self.create_logger(LEVELS.INFO)
        self.warning = self.create_logger(LEVELS.WARNING)
        self.error = self.create_logger(LEVELS.ERROR)
        self.audit = self.create_logger(LEVELS.AUDIT)
        self.alert = self.create_logger(LEVELS.ALERT)

    def __call__(self, message):
        self.debug(message)



import re
import datetime


TIMEDELTA_REGEX = (
    r"((?P<days>-?\d+)d)?"
    r"((?P<hours>-?\d+)h)?"
    r"((?P<minutes>-?\d+)m)?"
    r"((?P<seconds>-?\d+)s)?"
)
TIMEDELTA_PATTERN = re.compile(TIMEDELTA_REGEX, re.IGNORECASE)

# based on:
# https://gist.github.com/santiagobasulto/698f0ff660968200f873a2f9d1c4113c#file-parse_timedeltas-py
def parse_delta(delta: str) -> datetime.timedelta:
    """
    Parses a human readable timedelta (3d5h19m) into a datetime.timedelta.

    Delta includes:
    * Xd days
    * Xh hours
    * Xm minutes
    * Xs seconds

    Values can be negative following timedelta's rules. Eg: -5h-30m
    """
    match = TIMEDELTA_PATTERN.match(delta)
    if match:
        parts = {k: int(v) for k, v in match.groupdict().items() if v}
        return datetime.timedelta(**parts)
    return datetime.timedelta(seconds=0)



import os
import re
import datetime
from typing import Any, Union, List, Dict
from ..formats.json import parse
from ...errors import ValidationError
from ..formats.dictset.display import ascii_table


DEFAULT_MIN = -9223372036854775808
DEFAULT_MAX = 9223372036854775807
VALID_BOOLEAN_VALUES = {"true", "false", "on", "off", "yes", "no", "0", "1"}
CVE_REGEX = re.compile("cve|CVE-[0-9]{4}-[0-9]{4,}")


def is_boolean(**kwargs):
    def _inner(value: Any) -> bool:
        """boolean"""
        return str(value).lower() in VALID_BOOLEAN_VALUES

    return _inner


def is_cve(**kwargs):
    def _inner(value):
        """cve"""
        return CVE_REGEX.match(str(value))

    return _inner


def is_date(**kwargs):
    DATE_SEPARATORS = {"-", "\\", "/", ":"}
    # date validation at speed is hard, dateutil is great but really slow
    # this is as about as good as validating a string is a date, but isn't
    def _inner(value: Any) -> bool:
        """date"""
        try:
            if type(value).__name__ in ("datetime", "date", "time"):
                return True
            if type(value).__name__ == "str":
                if not value[4] in DATE_SEPARATORS:
                    return False
                if not value[7] in DATE_SEPARATORS:
                    return False
                if len(value) == 10:
                    # YYYY-MM-DD
                    datetime.date(*map(int, [value[:4], value[5:7], value[8:10]]))
                else:
                    if not value[10] == "T":
                        return False
                    if not value[13] in DATE_SEPARATORS:
                        return False
                    # YYYY-MM-DDTHH:MM
                    datetime.datetime(
                        *map(  # type:ignore
                            int,
                            [
                                value[:4],
                                value[5:7],
                                value[8:10],
                                value[11:13],
                                value[14:16],
                            ],
                        )
                    )
            return True
        except (ValueError, TypeError):
            return False

    return _inner


def is_list(**kwargs):
    def _inner(value: Any) -> bool:
        """list"""
        return isinstance(value, (list, set))

    return _inner


def is_null(**kwargs):
    def _inner(value: Any) -> bool:
        """nullable"""
        return (value is None) or (value == "") or (value == [])

    return _inner


def is_numeric(**kwargs):

    mn = kwargs.get("min") or DEFAULT_MIN
    mx = kwargs.get("max") or DEFAULT_MAX

    def _inner(value: Any) -> bool:
        """numeric"""
        try:
            n = float(value)
        except (ValueError, TypeError):
            return False
        return mn <= n <= mx

    return _inner


def is_string(**kwargs):
    regex = None
    pattern = kwargs.get("format")
    if pattern:
        regex = re.compile(pattern)

    def _inner(value: Any) -> bool:
        """string"""
        if pattern is None:
            return type(value).__name__ == "str"
        else:
            return regex.match(str(value))

    return _inner


def is_valid_enum(**kwargs):
    symbols = kwargs.get("symbols", set())

    def _inner(value: Any) -> bool:
        """enum"""
        return value in symbols

    return _inner


def other_validator(**kwargs):
    def _inner(value: Any) -> bool:
        """other"""
        return True

    return _inner


"""
Create dictionaries to look up the type validators
"""
VALIDATORS = {
    "date": is_date,
    "nullable": is_null,
    "other": other_validator,
    "list": is_list,
    "array": is_list,
    "enum": is_valid_enum,
    "numeric": is_numeric,
    "string": is_string,
    "boolean": is_boolean,
    "cve": is_cve,
}


class Schema:
    def __init__(self, definition: Union[str, List[Dict[str, Any]], dict]):
        """
        Tests a dictionary against a schema to test for conformity.
        Schema definition is similar to - but not the same as - avro schemas

        Paramaters:
            definition: dictionary or string
                A dictionary, a JSON string of a dictionary or the name of a
                JSON file containing a schema definition
        """
        # typing system struggles to understand what is happening here

        # if we have a schema as a string, load it into a dictionary
        if isinstance(definition, str):
            if os.path.exists(definition):  # type:ignore
                definition = parse(open(definition, mode="r").read())  # type:ignore
            else:
                definition = parse(definition)  # type:ignore

        if isinstance(definition, dict):
            if definition.get("fields"):  # type:ignore
                definition = definition["fields"]  # type:ignore

        try:
            # read the schema and look up the validators
            self._validators = {  # type:ignore
                item.get("name"): self._get_validators(  # type:ignore
                    item["type"],  # type:ignore
                    symbols=item.get("symbols"),  # type:ignore
                    min=item.get("min"),  # type:ignore
                    max=item.get("max"),  # type:ignore
                    format=item.get("format"),  # type:ignore
                )  # type:ignore
                for item in definition  # type:ignore
            }

        except KeyError:
            raise ValueError(
                "Invalid type specified in schema - valid types are: string, numeric, date, boolean, nullable, list, enum"
            )
        if len(self._validators) == 0:
            raise ValueError("Invalid schema specification")

    def _get_validators(self, type_descriptor: Union[List[str], str], **kwargs):
        """
        For a given type definition (the ["string", "nullable"] bit), return
        the matching validator functions (the _is_x ones) as a list.
        """
        if not type(type_descriptor).__name__ == "list":
            type_descriptor = [type_descriptor]  # type:ignore
        validators: List[Any] = []
        for descriptor in type_descriptor:
            validators.append(VALIDATORS[descriptor](**kwargs))
        return validators

    def _field_validator(self, value, validators: set) -> bool:
        """
        Execute a set of validator functions (the _is_x) against a value.
        Return True if any of the validators are True.
        """
        return any([True for validator in validators if validator(value)])

    def validate(self, subject: dict = {}, raise_exception=False) -> bool:
        """
        Test a dictionary against the Schema

        Parameters:
            subject: dictionary
                The dictionary to test for conformity
            raise_exception: boolean (optional, default False)
                If True, when the subject doesn't conform to the schema a
                ValidationError is raised

        Returns:
            boolean, True is subject conforms

        Raises:
            ValidationError
        """
        result = True
        self.last_error = ""

        for key, value in self._validators.items():
            if not self._field_validator(
                subject.get(key), self._validators.get(key, [other_validator])
            ):
                result = False
                for v in value:
                    self.last_error += f"'{key}' (`{subject.get(key)}`) did not pass `{v.__doc__}` validator.\n"
        if raise_exception and not result:
            raise ValidationError(
                f"Record does not conform to schema - {self.last_error}. "
            )
        return result

    def __call__(self, subject: dict = {}, raise_exception=False) -> bool:
        """
        Alias for validate
        """
        return self.validate(subject=subject, raise_exception=raise_exception)

    def __str__(self):
        retval = []
        for key, value in self._validators.items():
            val = [str(v).split(".")[0].split(" ")[1] for v in value]
            val = ",".join(val)
            retval.append({"field": key, "type": val})

        return ascii_table(retval)



"""
Advanced Error Trace Formatting

Adapted From:
https://github.com/willmcgugan/rich/blob/master/rich/traceback.py
Copyright 2020 Will McGugan
MIT Licence

#nodoc - don't add to the documentation wiki
"""
import os.path
import pathlib
import sys
from typing import Dict, List, Optional, Any, Generator
from dataclasses import dataclass, field
from traceback import walk_tb

LINE_LENGTH = 120
BAR = " " * LINE_LENGTH


@dataclass
class Frame:
    filename: str
    lineno: int
    name: str
    line: str = ""
    locals: Optional[Dict[str, Any]] = None


@dataclass
class Stack:
    exc_type: str
    exc_value: str
    is_cause: bool = False
    frames: List[Frame] = field(default_factory=list)


def wrap_text(text, line_len):
    from textwrap import fill

    def _inner(text):
        for line in text.splitlines():
            yield fill(line, line_len)

    return "\n".join(list(_inner(text)))


def bar_label(label):
    if len(label) > 0:
        center = ("**" + label + "**").center(LINE_LENGTH)
        return center.replace(" ", "=").replace("*", " ")
    else:
        return "-" * LINE_LENGTH


def _build_error_stack():
    exc_type, exc_value, traceback = sys.exc_info()

    # if there was no error, just return
    if exc_type is None:
        return []

    stacks = []
    is_cause = False

    while True:
        stack = Stack(
            exc_type=str(exc_type.__name__), exc_value=str(exc_value), is_cause=is_cause
        )

        stacks.append(stack)
        append = stack.frames.append

        for frame_summary, line_no in walk_tb(traceback):
            filename = frame_summary.f_code.co_filename
            filename = os.path.abspath(filename) if filename else "?"
            frame = Frame(
                filename=filename,
                lineno=line_no,
                name=frame_summary.f_code.co_name,
                locals={key: value for key, value in frame_summary.f_locals.items()},
            )
            append(frame)

        cause = getattr(exc_value, "__cause__", None)
        if cause and cause.__traceback__:
            exc_type = cause.__class__
            exc_value = cause
            traceback = cause.__traceback__
            if traceback:
                is_cause = True
                continue

        cause = exc_value.__context__
        if (
            cause
            and cause.__traceback__
            and not getattr(exc_value, "__suppress_context__", False)
        ):
            exc_type = cause.__class__
            exc_value = cause
            traceback = cause.__traceback__
            if traceback:
                is_cause = False
                continue
        # No cover, code is reached but coverage doesn't recognize it.
        break
    return stacks


def _render_locals(locals):
    if locals:
        yield bar_label("locals")
        max_label_len = max([len(key) for key, value in locals.items()])
        for key, value in locals.items():
            yield f"  {key:<{max_label_len}} : {value}"


def _read_from_code(filename: str, line: int, extend_by: int) -> Generator:
    try:
        with open(filename, "rt", encoding="utf-8") as code_file:
            code = code_file.read()

        lines = code.splitlines()
        start_line = max(line - extend_by, 0)
        end_line = min(line + extend_by + 1, len(lines) + 1)

        path = pathlib.Path(filename)

        yield bar_label(path.stem + path.suffix)
        for line_number in range(start_line, end_line):
            prefix = ">" if line_number == line else " "
            yield f"{prefix}{line_number:4d} | {lines[line_number-1]}"
            line_number += 1
    except:
        return ""


def _render_error_stack():
    stack = _build_error_stack()
    for item in stack:
        for frame in item.frames:
            yield f"{frame.filename}:{frame.lineno} in {frame.name}()"
            if frame.filename.startswith("<"):
                continue
            yield from _render_locals(frame.locals)
            yield from _read_from_code(
                filename=frame.filename, line=frame.lineno, extend_by=3
            )

            yield bar_label("")
            yield ""


def render_error_stack():
    s = [wrap_text(line, LINE_LENGTH) for line in _render_error_stack()]
    return "\n".join(s)



"""
Diablo: Python Graph Library

(C) 2021 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import types
from pathlib import Path
from .graph import Graph
from .traverse import Traverse
from ..json import parse
from ....errors import MissingDependencyError
from pydantic import BaseModel  # type:ignore


class EdgeModel(BaseModel):
    source: str
    target: str
    relationship: str


class NodeModel(BaseModel):
    nid: str
    display_name: str
    node_type: str
    attributes: dict = {}


def walk(graph, nids=None):
    """
    Begin a traversal by selecting the matching nodes.

    Parameters:
        *nids: strings
            the identity(s) of the node(s) to select

    Returns:
        A Diablo instance
    """
    if nids:
        nids = _make_a_list(nids)
        if len(nids) > 0:
            return Traverse(graph=graph, active_nodes=nids)
    else:
        return Traverse(graph, set())


def read_graphml(graphml_file: str):
    """
    Load a GraphML file into a Diablo Graph

    Parameters:
        graphml_file: string
            The GraphML file to load

    Returns:
        Graph
    """
    try:
        import xmltodict  # type:ignore
    except ImportError:  # pragma: no cover
        raise MissingDependencyError(
            "`xmltodict` is missing, please install or include in requirements.txt"
        )

    with open(graphml_file, "r") as fd:
        xml_dom = xmltodict.parse(fd.read())

    g = Graph()

    # load the keys
    keys = {}
    for key in xml_dom["graphml"].get("key", {}):
        keys[key["@id"]] = key["@attr.name"]

    g._nodes = {}
    # load the nodes
    for node in xml_dom["graphml"]["graph"].get("node", {}):
        data = {}
        skip = False
        for key in g._make_a_list(node.get("data", {})):
            try:
                data[keys[key["@key"]]] = key.get("#text", "")
            except KeyError:
                skip = True
        if not skip:
            g.add_node(node["@id"], data)

    g._edges = {}
    for edge in xml_dom["graphml"]["graph"].get("edge", {}):
        data = {}
        source = edge["@source"]
        target = edge["@target"]
        for key in g._make_a_list(edge.get("data", {})):
            data[keys[key["@key"]]] = key["#text"]
        if source not in g._edges:
            g._edges[source] = []
        g.add_edge(source, target, data.get("relationship"))  # type:ignore

    return g


def _load_node_file(path: Path):
    """load the node information from a file"""
    nodes = []
    with open(path, "r", encoding="utf8") as node_file:
        for line in node_file:
            node = parse(line)
            nodes.append(
                (
                    node["nid"],
                    node["attributes"],
                )
            )
    results = {n: a for n, a in nodes}
    return results


def _load_edge_file(path: Path):
    """load the edge information from a file"""
    edges = []
    with open(path, "r", encoding="utf8") as edge_file:
        for line in edge_file:
            node = parse(line)
            edges.append(
                (
                    node["source"],
                    node["target"],
                    node["relationship"],
                )
            )
    results: dict = {s: [] for s, t, r in edges}
    for s, t, r in edges:
        results[s].append(
            (
                t,
                r,
            )
        )
    return results


def load(path: str):
    """
    Load a saved Diablo graph.

    Parameters:
        path: string
            The path to the folder containing the Graph files

    Returns:
        Graph
    """
    g = Graph()
    graph_path = Path(path)
    g._nodes = _load_node_file(graph_path / "nodes.jsonl")
    g._edges = _load_edge_file(graph_path / "edges.jsonl")
    return g


def _make_a_list(obj):
    """internal helper method"""
    if isinstance(obj, (set, list, types.GeneratorType)):
        return obj
    return [obj]



from ...formats.json import parse
from ....errors import MissingDependencyError

try:
    import xmltodict  # type:ignore
except ImportError:
    xmltodict = None


def json_parser(ds):
    """parse each line in the file to a dictionary"""
    yield from map(parse, ds)


def pass_thru_parser(ds):
    """just pass it through"""
    yield from ds


def block_parser(ds):
    """each blob is read as a block"""
    if isinstance(ds, str):
        return ds
    yield "\n".join([r for r in ds])  # pragma: no cover


def xml_parser(ds):
    if not xmltodict:  # pragma: no cover
        raise MissingDependencyError(
            "`xmltodict` is missing, please install or include in requirements.txt"
        )
    yield from map(xmltodict.parse, ds)



"""
Diablo: Python Graph Library

(C) 2021 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Callable


class Traverse:

    __slots__ = ("graph", "_active_nodes", "_active_nodes_cache")

    def __init__(self, graph, active_nodes: set = None):
        """
        Graph Traversal

        Parameters:
            graph: Graph
            active_nodes: Iterable
        """
        self.graph = graph
        self._active_nodes_cache = None

        # ensure it is a set
        # - the collection active nodes is immutable
        # - sets are faster for look ups
        if not active_nodes:
            self._active_nodes = set()
        elif type(active_nodes).__name__ == "set":
            self._active_nodes = active_nodes
        else:
            self._active_nodes = set(active_nodes)

    def follow(self, *relationships):
        """
        Traverses a graph by following edges from the active nodes with
        a relationship on the list of relationships.

        Parameters:
            relationsips: strings
                traverses node following edges with the stated relationship

        Returns:
            Diablo
        """
        active_nodes = []

        for node in self._active_nodes:
            active_nodes += [
                t for (s, t, r) in self.graph.outgoing_edges(node) if r in relationships
            ]
        return Traverse(graph=self.graph, active_nodes=active_nodes)

    def select(self, filter: Callable):
        """
        Filters the active nodes by a function.

        Parameters:
            filter: Callable
                node attribute name to filter on

        Returns:
            Diablo
        """
        active_nodes = {
            nid for nid, attrib in self.active_nodes(data=True) if filter(attrib)
        }
        return Traverse(graph=self.graph, active_nodes=active_nodes)

    def has(self, key, value):
        """
        Filters the active nodes by a key/value match
        """
        active_nodes = [
            nid
            for nid, attrib in self.active_nodes(data=True)
            if attrib.get(key) == value
        ]
        return Traverse(graph=self.graph, active_nodes=active_nodes)

    def values(self, key):
        return list(
            {
                attrib[key]
                for nid, attrib in self.active_nodes(data=True)
                if key in attrib
            }
        )

    def active_nodes(self, data=False):
        if not data:
            return self._active_nodes
        if not self._active_nodes_cache:
            self._active_nodes_cache = [
                (nid, self.graph[nid]) for nid in self._active_nodes
            ]
        return self._active_nodes_cache

    def list_relationships(self):
        relationships = []
        for node in self._active_nodes:
            relationships += {r for (s, t, r) in self.graph.outgoing_edges(node)}
        return set(relationships)

    def __repr__(self):
        return f"Graph - {len(list(self.graph.nodes()))} nodes ({len(self._active_nodes)} selected), {len(list(self.graph.edges()))} edges"

    def __len__(self):
        return len(self._active_nodes)



"""
Dictionary Validation Operator

Checks a dictionary against a schema.
"""
from ..flows.internals.base_operator import BaseOperator
from ..data.validator import Schema  # type:ignore
from typing import Any


class ValidatorOperator(BaseOperator):
    def __init__(self, schema: Any):
        self.validator = Schema(schema)
        self.invalid_records = 0
        super().__init__()

    def execute(self, data={}, context={}):
        valid = self.validator(subject=data)
        if not valid:
            self.errors += 1
            return None
        else:
            return data, context



import os
import datetime
import traceback
from typing import Any
from .simple_writer import SimpleWriter
from .internals.blob_writer import BlobWriter
from ...utils import paths
from ...data.formats import json
from ...logging import get_logger


class BatchWriter(SimpleWriter):
    def __init__(
        self,
        *,
        dataset: str,
        format: str = "zstd",
        date: Any = None,
        frame_id: str = None,
        **kwargs,
    ):
        """
        The batch data writer to writes data records into blobs. Batches
        are written into timestamped folders called Partitions.

        Parameters:
            dataset: string (optional)
                The name of the dataset - this is used to map to a path
            schema: mabel.validator.Schema (optional)
                Schema used to test records for conformity, default is no
                schema and therefore no validation
            format: string (optional)
                - text: raw text lines
                - jsonl: raw json lines
                - lzma: lzma compressed json lines
                - zstd: zstandard compressed json lines (default)
                - parquet: Apache Parquet
            date: date or string (optional)
                A date, a string representation of a date to use for
                creating the dataset. The default is today's date
            blob_size: integer (optional)
                The maximum size of blobs, the default is 64Mb
            inner_writer: BaseWriter (optional)
                The component used to commit data, the default writer is the
                NullWriter
            frame_id: string (optional)

            raw_path: boolean (optional)
                Don't automatically add any date parts to dataset names
            index_on: collection (optional)
                Index on these columns, the default is to not index

        Note:
            Different inner_writers may take or require additional parameters.
        """
        # call this now, we need this value before we call the base class init
        self.batch_date = self._get_writer_date(date)
        # because jobs can be split, allow a frame_id to be passed in
        if not frame_id:
            frame_id = BatchWriter.create_frame_id()

        self.dataset = dataset
        if "{date" not in self.dataset and not kwargs.get("raw_path", False):
            self.dataset += "/{datefolders}"
        self.dataset = paths.build_path(
            self.dataset + "/" + frame_id,  # type:ignore
            self.batch_date,
        )

        kwargs["raw_path"] = True  # we've just added the dates
        kwargs["format"] = format
        kwargs["dataset"] = self.dataset

        super().__init__(**kwargs)

        self.dataset = paths.build_path(self.dataset, self.batch_date)

        # create the writer
        self.blob_writer = BlobWriter(**kwargs)

    def finalize(self):
        final = super().finalize()

        ex = traceback.format_exc()
        if ex != "NoneType: None\n":
            get_logger().debug(
                f"Error found in the stack, not marking frame as complete."
            )
            return -1

        completion_path = self.blob_writer.inner_writer.filename
        completion_path = os.path.split(completion_path)[0] + "/frame.complete"
        status = {"records": self.records}
        flag = self.blob_writer.inner_writer.commit(
            byte_data=json.serialize(status, as_bytes=True),
            override_blob_name=completion_path,
        )
        get_logger().debug(f"Frame completion file `{flag}` written")
        return final

    @staticmethod
    def create_frame_id():
        return datetime.datetime.now().strftime("as_at_%Y%m%d-%H%M%S")



# no-maintain-checks
from moz_sql_parser import parse  # type:ignore
from ..reader import Reader
from ....logging import get_logger
from ...formats import json
from functools import lru_cache
import re

# https://codereview.stackexchange.com/a/248421
_special_regex_chars = {ch: "\\" + ch for ch in ".^$*+?{}[]|()\\"}


@lru_cache(4)
def _sql_like_fragment_to_regex(fragment):
    # https://codereview.stackexchange.com/a/36864/229677
    safe_fragment = "".join([_special_regex_chars.get(ch, ch) for ch in fragment])
    return re.compile("^" + safe_fragment.replace("%", ".*?").replace("_", ".") + "$")


def _div(x, y):
    return x / y


def _like(x, y):
    return _sql_like_fragment_to_regex(y).match(str(x))


import operator

# functions which implement the Operators
OPERATORS = {
    "eq": "==",
    "neq": "!=",
    "lt": "<",
    "gt": ">",
    "lte": ">=",
    "gte": "<=",
    "like": "like",
    "add": operator.add,
    "mul": operator.mul,
    "div": _div,
    "sub": operator.sub,
}


def _get_operand(operand):
    if not isinstance(operand, dict):
        # strings are field names
        #        if isinstance(operand, str):
        #            return row[operand]
        # otherwise it's a constant
        return operand
    # string constants are 'literals'
    if "literal" in operand:
        return operand["literal"]
    # some values are handled unusually
    #    value = [row.get(k) for k, v in operand.items() if k in ["timestamp", "text"]]
    #    if len(value):
    #        return value.pop()
    # if we're here, the operand is probably a function
    return _build_filters(operand)


def _build_filters(where_object):
    # None is None
    filters = []
    if where_object is None:
        return None
    # the Operator is the head of the dictionary
    operator = list(where_object.keys())[0]
    if operator in ("and"):
        for predicate in where_object[operator]:
            filters.append(_build_filters(predicate))
        return filters
    if operator in ("or"):
        for predicate in where_object[operator]:
            filters.append([_build_filters(predicate)])
        return filters

    # the operands are the values in for the key
    operands = where_object.get(operator)
    # if the operands are a list, it's the left and right operands
    if isinstance(operands, list):
        left_operand, right_operand = operands[0], operands[1]

    # we build a function, load some of the values and pass it to the Reader
    return (left_operand, OPERATORS.get(operator), _get_operand(right_operand))


class SqlReader:
    def __init__(self, sql_statement: str, **kwargs):
        """
        THIS IS ALPHA - interface and features subject to change

        Use basic SQL queries to filter Reader.

        Parameters:
            sql_statement: string
            kwargs: parameters to pass to the Reader

        Note:
            `select` is taken from SQL SELECT
            `dataset` is taken from SQL FROM
            `where` is taken from SQL WHERE
        """
        get_logger().warning(
            "SQL Reader is Alpha - interface and features subject to change"
        )

        sql = parse(sql_statement)
        get_logger().debug(json.serialize(sql))

        field_list = sql.get("select")
        fields = []
        if isinstance(field_list, str):
            if field_list == "*":
                field_list = {"value": "*"}
            field_list = [field_list]
        elif not isinstance(field_list, list):
            field_list = [field_list]
        fields = [field.get("value") for field in field_list]

        table = sql.get("from").replace(".", "/")

        # where_statement = _build_where_function(sql.get("where"))
        filters = _build_filters(sql.get("where"))
        print(">>", filters)

        thread_count = 4

        self.reader = Reader(
            #            thread_count=thread_count,
            select=fields,
            filters=filters,
            dataset=table,
            **kwargs,
        )

    def __iter__(self):
        return self

    def __next__(self):
        # deepcode ignore unguarded~next~call: error should bubble
        return next(self.reader)



import time
import datetime
import threading
from pydantic import BaseModel  # type:ignore
from typing import Any, Union
from .simple_writer import SimpleWriter
from .internals.writer_pool import WriterPool
from ...utils import paths
from ...logging import get_logger


class StreamWriter(SimpleWriter):
    def __init__(
        self,
        *,
        dataset: str,
        format: str = "zstd",
        idle_timeout_seconds: int = 30,
        writer_pool_capacity: int = 5,
        **kwargs,
    ):
        """
        Create a Data Writer to write data records into partitions.

        Parameters:
            dataset: string (optional)
                The name of the dataset - this is used to map to a path
            schema: mabel.validator.Schema (optional)
                Schema used to test records for conformity, default is no
                schema and therefore no validation
            format: string (optional)
                - jsonl: raw json lines
                - lzma: lzma compressed json lines
                - zstd: zstandard compressed json lines (default)
                - parquet: Apache Parquet
            idle_timeout_seconds: integer (optional)
                The number of seconds to wait before evicting writers from the
                pool for inactivity, default is 30 seconds
            writer_pool_capacity: integer (optional)
                The number of writers to leave in the writers pool before
                writers are evicted for over capacity, default is 5
            blob_size: integer (optional)
                The maximum size of blobs, the default is 32Mb
            inner_writer: BaseWriter (optional)
                The component used to commit data, the default writer is the
                NullWriter

        Note:
            Different inner_writers may take or require additional parameters.
        """
        if kwargs.get("date"):
            get_logger().warning("Cannot specify a `date` for the StreamWriter.")

        # add the values to kwargs
        kwargs["format"] = format
        kwargs["dataset"] = dataset
        self.dataset = dataset

        super().__init__(**kwargs)

        self.idle_timeout_seconds = idle_timeout_seconds

        # we have a pool of writers of size maximum_writers
        self.writer_pool_capacity = writer_pool_capacity
        self.writer_pool = WriterPool(pool_size=writer_pool_capacity, **kwargs)

        # establish the background thread responsible for the pool
        self.thread = threading.Thread(target=self.pool_attendant)
        self.thread.name = "mabel-writer-pool-attendant"
        self.thread.daemon = True
        self.thread.start()

    def append(self, record: Union[dict, BaseModel]):
        """
        Append a new record to the Writer

        Parameters:
            record: dictionary
                The record to append to the Writer

        Returns:
            integer
                The number of records in the current blob
        """
        # get the appropritate writer from the pool and append the record
        # the writer identity is the base of the path where the partitions
        # are written.

        # Check the new record conforms to the schema
        # unlike the batch writer, we don't want to bail out if we have a
        # problem here, instead we're going to save the file to a BACKOUT
        # partition

        identity = paths.date_format(self.dataset_template, datetime.date.today())

        if isinstance(record, BaseModel):
            record = record.dict()
        elif self.schema and not self.schema.validate(
            subject=record, raise_exception=False
        ):
            identity += "/BACKOUT/"
            get_logger().warning(
                f"Schema Validation Failed ({self.schema.last_error}) - message being written to {identity}"
            )

        with threading.Lock():
            blob_writer = self.writer_pool.get_writer(identity)
            return blob_writer.append(record)

    def finalize(self):
        with threading.Lock():
            for blob_writer_identity in self.writer_pool.writers:
                try:
                    get_logger().debug(
                        f"Removing {blob_writer_identity} from the writer pool during finalization"
                    )
                    self.writer_pool.remove_writer(blob_writer_identity)
                except Exception as err:
                    get_logger().debug(
                        f"Error finalizing `{blob_writer_identity}`, {type(err).__name__} - {err}"
                    )
        return super().finalize()

    def pool_attendant(self):
        """
        Writer Pool Management
        """
        while True:
            with threading.Lock():
                # search for pool occupants who haven't had a write recently
                for blob_writer_identity in self.writer_pool.get_stale_writers(
                    self.idle_timeout_seconds
                ):
                    get_logger().debug(
                        f"Evicting {blob_writer_identity} from the writer pool due to inactivity - limit is {self.idle_timeout_seconds} seconds"
                    )
                    self.writer_pool.remove_writer(blob_writer_identity)
                # if we're over capacity, evict the LRU writers
                for (
                    blob_writer_identity
                ) in self.writer_pool.nominate_writers_to_evict():
                    get_logger().debug(
                        f"Evicting {blob_writer_identity} from the writer pool due the pool being over its {self.writer_pool_capacity} capacity"
                    )
                    self.writer_pool.remove_writer(blob_writer_identity)
            time.sleep(1)



# nodoc - don't add to the documentation wiki
"""
This exception should be used when a lazy import fails
"""
from .base_exception import BaseException


class MissingDependencyError(BaseException):
    pass



import io
from ...data.readers.internals.base_inner_reader import BaseInnerReader


class DiskReader(BaseInnerReader):
    def __init__(self, **kwargs):
        """
        File System Reader
        """
        super().__init__(**kwargs)

    def get_blobs_at_path(self, path):
        if path.exists():
            blobs = path.rglob("*")
            return [
                blob.as_posix()
                for blob in blobs
                if blob.is_file()
                and (
                    blob.suffix in self.VALID_EXTENSIONS
                    or blob.stem in self.VALID_EXTENSIONS
                )
            ]
        return []

    def get_blob_stream(self, blob_name: str) -> io.IOBase:
        with open(blob_name, "rb") as f:
            io_stream = io.BytesIO(f.read())
            return io_stream



"""
File Bin
"""
import time
from .base_bin import BaseBin
import os


class FileBin(BaseBin):
    def __init__(self, bin_name: str, path: str):
        self.path = path
        self.name = bin_name
        os.makedirs(self.path, exist_ok=True)

    def __call__(self, record: str, id_: str = ""):

        os.makedirs(f"{self.path}/{self._date_part()}", exist_ok=True)
        filename = f"{self.path}/{self._date_part()}/{id_}{time.time_ns()}.txt"
        with open(filename, "w") as file:
            file.write(record)
        return filename



import os
from ...data.writers.internals.base_inner_writer import BaseInnerWriter
from ...logging import get_logger
from ...utils import paths


class DiskWriter(BaseInnerWriter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def commit(self, byte_data, override_blob_name=None):

        try:
            # if we've been given the filename, use that, otherwise get the
            # name from the path builder
            if override_blob_name:
                blob_name = override_blob_name
            else:
                blob_name = self._build_path()

            bucket, path, stem, ext = paths.get_parts(blob_name)

            os.makedirs(bucket + "/" + path, exist_ok=True)
            with open(blob_name, mode="wb") as file:
                file.write(byte_data)

            return blob_name
        except Exception as err:  # pragma: no cover
            get_logger().error(f"Problem saving blob to disk {type(err).__name__}")
            raise err



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import StreamWriter  # type:ignore
from ...adapters.disk import DiskWriter


class DiskStreamWriterOperator(BaseWriterOperator):
    def __init__(self, **kwargs):

        super().__init__(writer=StreamWriter, inner_writer=DiskWriter, **kwargs)



import threading
import tempfile
import os
from typing import Any
from ...formats.json import serialize
from ....logging import get_logger
from ....utils.paths import get_parts
from ....utils import safe_field_name
from ....index.index import IndexBuilder
from ....errors import MissingDependencyError


BLOB_SIZE = 64 * 1024 * 1024  # 64Mb, 16 files per gigabyte
BUFFER_SIZE = BLOB_SIZE  # buffer in memory really
SUPPORTED_FORMATS_ALGORITHMS = ("jsonl", "lzma", "zstd", "parquet", "text")


class BlobWriter(object):

    # in som failure scenarios commit is called before __init__, so we need to define
    # this variable outside the __init__.
    bytes_in_blob = 0

    def __init__(
        self,
        *,  # force params to be named
        inner_writer=None,  # type:ignore
        blob_size: int = BLOB_SIZE,
        format: str = "zstd",
        **kwargs,
    ):

        self.indexes = kwargs.get("index_on", [])

        self.format = format
        self.maximum_blob_size = blob_size

        if format not in SUPPORTED_FORMATS_ALGORITHMS:
            raise ValueError(
                f"Invalid format `{format}`, valid options are {SUPPORTED_FORMATS_ALGORITHMS}"
            )

        kwargs["format"] = format
        self.inner_writer = inner_writer(**kwargs)  # type:ignore

        self._open_blob()

    def append(self, record: dict = {}):
        # serialize the record
        if self.format == "text":
            serialized = str(record).encode() + b"\n"
        else:
            serialized = serialize(record, as_bytes=True) + b"\n"  # type:ignore

        # add the columns to the index
        for column in self.indexes:
            self.index_builders[column].add(self.records_in_blob, record)

        # the newline isn't counted so add 1 to get the actual length
        # if this write would exceed the blob size, close it so another
        # blob will be created
        self.bytes_in_blob += len(serialized) + 1
        if self.bytes_in_blob > self.maximum_blob_size and self.records_in_blob > 0:
            self.bytes_in_blob -= len(serialized) + 1
            self.commit()
            self._open_blob()

        # write the record to the file
        self.file.write(serialized)
        self.records_in_blob += 1

        return self.records_in_blob

    def commit(self):

        committed_blob_name = ""

        if self.bytes_in_blob > 0:
            with threading.Lock():
                try:
                    self.file.flush()
                    self.file.close()
                except ValueError:
                    pass

                if self.format == "parquet":
                    try:
                        from pyarrow import json as js  # type:ignore
                        import pyarrow.parquet as pq  # type:ignore
                    except ImportError as err:  # pragma: no cover
                        raise MissingDependencyError(
                            "`pyarrow` is missing, please install or includein requirements.txt"
                        )

                    table = js.read_json(self.file_name)
                    pq.write_table(
                        table, self.file_name + ".parquet", compression="ZSTD"
                    )
                    self.file_name += ".parquet"

                with open(self.file_name, "rb") as f:
                    byte_data = f.read()

                committed_blob_name = self.inner_writer.commit(
                    byte_data=byte_data, override_blob_name=None
                )

                for column in self.indexes:
                    index = self.index_builders[column].build()

                    bucket, path, stem, suffix = get_parts(committed_blob_name)
                    index_name = (
                        bucket
                        + "/"
                        + path
                        + "_SYS."
                        + stem
                        + "."
                        + safe_field_name(column)
                        + ".index"
                    )

                    index_stream = index._index
                    index_stream.seek(0)
                    committed_index_name = self.inner_writer.commit(
                        byte_data=index_stream.read(), override_blob_name=index_name
                    )

                if "BACKOUT" in committed_blob_name:
                    get_logger().warning(
                        f"{self.records_in_blob:n} failed records written to BACKOUT partition `{committed_blob_name}`"
                    )
                get_logger().debug(
                    {
                        "committed_blob": committed_blob_name,
                        "records": self.records_in_blob,
                        "raw_bytes": self.bytes_in_blob,
                        "committed_bytes": len(byte_data),
                    }
                )
                try:
                    os.remove(self.file_name)
                except ValueError:
                    pass

                self.bytes_in_blob = 0
                self.file_name = None

        return committed_blob_name

    def _open_blob(self):
        self.file_name = self._create_temp_file_name()
        self.file: Any = open(self.file_name, mode="wb", buffering=BUFFER_SIZE)
        if self.format == "lzma":
            import lzma

            self.file = lzma.open(self.file, mode="wb")
        if self.format == "zstd":
            import zstandard  # type:ignore

            self.file = zstandard.open(self.file_name, mode="wb")

        # create index builders
        self.index_builders = {}
        for column in self.indexes:
            self.index_builders[column] = IndexBuilder(column)

        self.bytes_in_blob = 0
        self.records_in_blob = 0

    def __del__(self):
        # this should never be relied on to save data
        self.commit()

    def _create_temp_file_name(self):
        """
        Create a tempfile, get the name and then deletes the tempfile.

        The behaviour of tempfiles is inconsistent between operating systems,
        this helps to ensure consistent behaviour.
        """
        file = tempfile.NamedTemporaryFile(prefix="mabel-", delete=True)
        file_name = file.name
        file.close()
        try:
            os.remove(file_name)
        except OSError:
            pass
        return file_name



"""
End Operator

Explicit marker for the end of a flow.

The EndOperator has no configuration options, it is currently implemented to
terminate flows - although this is mostly a passive activity, done by
returning 'None'.

The purpose of this Operator is three-fold:

- Makes it clear to human readers where flows end
- Prevents issues with the flow builder
- Provides opportunity in the future to introduce functionality at the 
  completion of a flow
"""
from ..flows.internals.base_operator import BaseOperator


class EndOperator(BaseOperator):
    def execute(self, data={}, context={}):
        # do nothing
        pass



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import BatchWriter
from ...adapters.disk import DiskWriter


class DiskBatchWriterOperator(BaseWriterOperator):
    def __init__(self, **kwargs):

        super().__init__(writer=BatchWriter, inner_writer=DiskWriter, **kwargs)



# Store the version here so:
# 1) we don't load dependencies by storing it in __init__.py
# 2) we can import it in setup.py for the same reason
__version__ = "0.4.61"

# nodoc - don't add to the documentation wiki



from .ipython import is_running_from_ipython


def safe_field_name(field_name):
    """strip all the non-alphanums from a field name"""
    import re

    pattern = re.compile("[^a-zA-Z0-9\_\-]+")
    return pattern.sub("", field_name)



"""
Base Inner Reader
"""
import os
import abc
import pathlib
import datetime
from io import IOBase
from typing import Iterable, Optional
from dateutil import parser
from ...formats import json
from ....utils import common, paths
from ....logging import get_logger
from ....errors import MissingDependencyError


def zstd_reader(stream, rows, all_rows):
    """
    Read zstandard compressed files
    """
    # zstandard should always be present
    import zstandard  # type:ignore

    with zstandard.open(stream, "r", encoding="utf8") as file:  # type:ignore
        for index, row in enumerate(file):
            if all_rows or index in rows:
                yield row


def lzma_reader(stream, rows, all_rows):
    """
    Read LZMA compressed files
    """
    # lzma should always be present
    import lzma

    with lzma.open(stream, "rb") as file:  # type:ignore
        for index, row in enumerate(file):
            if all_rows or index in rows:
                yield row


def parquet_reader(stream, rows, all_rows):
    """
    Read parquet formatted files
    """
    try:
        import pyarrow.parquet as pq  # type:ignore
    except ImportError:  # pragma: no cover
        raise MissingDependencyError(
            "`pyarrow` is missing, please install or include in requirements.txt"
        )
    table = pq.read_table(stream)
    for batch in table.to_batches():
        dict_batch = batch.to_pydict()
        for index in range(len(batch)):
            if all_rows or index in rows:
                yield json.serialize(
                    {k: v[index] for k, v in dict_batch.items()}
                )  # type:ignore


def text_reader(stream, rows, all_rows):
    """
    Default reader, assumes text format
    """
    text = stream.read().decode("utf8")  # type:ignore
    lines = text.splitlines()
    for index, row in enumerate(lines):
        if all_rows or index in rows:
            yield row


READERS = {".zstd": zstd_reader, ".lzma": lzma_reader, ".parquet": parquet_reader}


class BaseInnerReader(abc.ABC):

    VALID_EXTENSIONS = (
        ".txt",
        ".json",
        ".zstd",
        ".lzma",
        ".jsonl",
        ".csv",
        ".lxml",
        ".parquet",
        ".ignore",
        ".profile",
        ".index",
        ".bloom",
        ".complete",
    )

    def _extract_date_part(self, value):
        if isinstance(value, str):
            value = parser.parse(value)
        if isinstance(value, (datetime.date, datetime.datetime)):
            return datetime.date(value.year, value.month, value.day)
        return datetime.date.today()

    def _extract_as_at(self, path):
        parts = path.split("/")
        for part in parts:
            if part.startswith("as_at_"):
                return part
        return ""

    def __init__(self, **kwargs):
        self.dataset = kwargs.get("dataset")
        if self.dataset is None:
            raise ValueError("Readers must have the `dataset` parameter set")
        if not self.dataset.endswith("/"):
            self.dataset += "/"
        if "date" not in self.dataset and not kwargs.get("raw_path", False):
            self.dataset += "{datefolders}/"

        self.start_date = self._extract_date_part(kwargs.get("start_date"))
        self.end_date = self._extract_date_part(kwargs.get("end_date"))

        self.days_stepped_back = 0

    def step_back_a_day(self):
        """
        Steps back a day so data can be read from a previous day
        """
        self.days_stepped_back += 1
        self.start_date -= datetime.timedelta(days=1)
        self.end_date -= datetime.timedelta(days=1)
        return self.days_stepped_back

    def __del__(self):
        """
        Only here in case a helpful dev expects te base class to have it
        """
        pass

    @abc.abstractmethod
    def get_blobs_at_path(self, prefix=None) -> Iterable:
        pass

    @abc.abstractmethod
    def get_blob_stream(self, blob: str) -> IOBase:
        """
        Return a filelike object
        """
        pass

    def get_records(
        self, blob_name: str, rows: Optional[Iterable[int]] = None
    ) -> Iterable[str]:
        """
        Handle the different file formats.

        Handling here allows the blob stores to be pretty dumb, they
        just need to be able to store and recall blobs.

        Parameters:
            blob_name: string
                The name of the blob to read records from
            rows: list of integers (optional)
                The row numbers of the blob to retrieve records from, the
                default (None) is all of the rows. An empty set returns an
                empty Iterable from this method

        Returns:
            Iterable of the rows
        """
        get_logger().debug(f"Reading data from `{blob_name}`")
        # if the rows has been set but the set is empty, return empty list
        if isinstance(rows, (set, list)):
            if len(rows) == 0:
                return []
            rows = set(rows)
            all_rows = False
        else:
            all_rows = True
            rows = set([-1])

        path, ext = os.path.splitext(blob_name)
        stream = self.get_blob_stream(blob_name)

        yield from READERS.get(ext, text_reader)(stream, rows, all_rows)

    def get_list_of_blobs(self):

        blobs = []
        for cycle_date in common.date_range(self.start_date, self.end_date):
            # build the path name
            cycle_path = pathlib.Path(
                paths.build_path(path=self.dataset, date=cycle_date)
            )
            blobs += list(self.get_blobs_at_path(path=cycle_path))

        # remove any BACKOUT data
        blobs = [blob for blob in blobs if "BACKOUT" not in blob]

        # work out if there's an as_at part
        as_ats = {self._extract_as_at(blob) for blob in blobs if "as_at_" in blob}
        if as_ats:
            as_ats = sorted(as_ats)
            as_at = as_ats.pop()

            is_complete = lambda blobs: any(
                [blob for blob in blobs if as_at + "/frame.complete" in blob]
            )
            is_invalid = lambda blobs: any(
                [blob for blob in blobs if (as_at + "/frame.ignore" in blob)]
            )

            while not is_complete(blobs) or is_invalid(blobs):
                if not is_complete(blobs):
                    get_logger().warning(
                        f"Frame `{as_at}` is not complete - `frame.complete` file is not present - skipping this frame."
                    )
                if is_invalid(blobs):
                    get_logger().debug(
                        f"Frame `{as_at}` is invalid - `frame.ignore` file is present - skipping this frame."
                    )
                if len(as_ats) > 0:
                    as_at = as_ats.pop()
                else:
                    return []
            get_logger().debug(f"Reading from DataSet frame `{as_at}`")
            blobs = [
                blob
                for blob in blobs
                if (as_at in blob) and ("/frame.complete" not in blob)
            ]

        return sorted(blobs)



# no-maintain-checks
"""
This module is dervied from:

Cryno: Data Profiling Library
https://github.com/joocer/cryno

(C) 2021 Justin Joyce.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

┌──────────────────────────────────────────────────────────────────────────────┐
│          NOTE: BINNING HAS BEEN REMOVED AS BOTH SLOW AND INACCURATE          │
└──────────────────────────────────────────────────────────────────────────────┘

"""
from ..flows.internals.base_operator import BaseOperator
from ..data.formats import json
from ..data.formats.dictset.display import draw_histogram_bins
from ..index.bloom_filter import BloomFilter

ACCURATE_THRESHOLD = 5000
BLOOM_FILTER_SIZE = 2500000


class ProfileDataOperator(BaseOperator):
    def __init__(self, schema=None, inner_writer=None, **kwargs):
        """
        ProfileData Operator is designed to profile streaming data this is done
        with some approximation, especially when summarizations would require
        all of the values such as counting unique values or binning values.

        It starts summarizing from 5,000 values, small datasets will be more
        accurate. Some measures will be pretty much useless the closer it gets
        to 2,500,000 records.

        These limits are in place to reduce memory overheads, even with the
        approaches taken this is pretty memory hungry.
        """
        super().__init__(**kwargs)

        self.inner_writer = inner_writer
        self.fields = schema._validators

        self.summary = {}
        for field in self.fields:
            self.summary[field] = {
                "nulls": 0,
                "items": 0,
                "type": ProfileDataOperator.get_type(schema._validators[field]),
            }
        self.logger.warning(
            "`ProfileDataOperator` is an alpha feature and may change in the future."
        )

    def execute(self, data, context):

        for field in self.fields:
            field_summary = self.summary[field]
            field_summary["items"] += 1
            field_type = field_summary["type"]
            field_value = data.get(field)
            if field_value is None:
                field_summary["nulls"] += 1

            elif field_type == "numeric":
                if isinstance(field_value, str):
                    field_value = float(field_value)
                field_summary = ProfileDataOperator.profile_numeric_data(
                    field_summary, field_value
                )

            elif field_type == "date":
                # convert to epoch seconds and then just treat as a number
                from dateutil import parser

                field_value = int(parser.isoparse(field_value).timestamp())
                field_summary = ProfileDataOperator.profile_numeric_data(
                    field_summary, field_value
                )

            elif field_type == "string":
                field_summary = ProfileDataOperator.profile_string_data(
                    field_summary, field_value
                )

            elif field_type == "enum":
                """
                Profiler for enums (including boolean), just counts the
                number of values in each category
                """
                if field_summary.get("values") is None:
                    field_summary["values"] = {}
                field_summary["values"][f"{field_value}"] = (
                    field_summary["values"].get(f"{field_value}", 0) + 1
                )

        return data, context

    def finalize(self, context: dict = None):

        if not context:
            context = {}

        for k in [k for k in self.summary if self.summary[k]["type"] == "string"]:

            if len(self.summary[k]["unique_values"]) < ACCURATE_THRESHOLD:
                # get rid of the unique values list
                self.summary[k]["unique_values"] = len(self.summary[k]["unique_values"])
                del self.summary[k]["approx_unique_values"]
            else:
                del self.summary[k]["unique_values"]

            del self.summary[k]["bloom_filter"]

        # rebin all of the binned data into a set of 10 bins
        # new_bins = [
        #    (k, ProfileDataOperator.redistribute_bins(self.summary[k]["bins"], 10))
        #    for k in self.summary
        #    if "bins" in self.summary[k]
        # ]
        # for k, bins in new_bins:
        #    self.summary[k]["bins"] = bins

        # convert date fields back to dates
        date_fields = [k for k in self.summary if self.summary[k]["type"] == "date"]
        for date_field in date_fields:
            self.summary[date_field]["min"] = ProfileDataOperator.date_from_epoch(
                self.summary[date_field].get("min", 0)
            )
            self.summary[date_field]["max"] = ProfileDataOperator.date_from_epoch(
                self.summary[date_field].get("max", 0)
            )
            # new_bins = {}  # type:ignore
            # for bound in self.summary[date_field].get("bins", []):
            #    bottom, top = bound
            #    bottom = ProfileDataOperator.date_from_epoch(bottom)
            #    top = ProfileDataOperator.date_from_epoch(top)
            #    new_bins[(bottom, top)] = self.summary[date_field][  # type:ignore
            #        "bins"
            #    ][bound]
            # self.summary[date_field]["bins"] = new_bins
            if self.summary[date_field].get("mean"):
                del self.summary[date_field]["mean"]
            if self.summary[date_field].get("cumsum"):
                del self.summary[date_field]["cumsum"]

        # change the bin names to strings from tuples
        # for k in [k for k in self.summary if "bins" in self.summary[k]]:
        #    new_bins = {}  # type:ignore
        #    old_bins = self.summary[k]["bins"]
        #    for ranges, v in old_bins.items():
        #        bottom, top = ranges
        #        label = f"{ProfileDataOperator.short_form(bottom)} to {ProfileDataOperator.short_form(top)}"
        #        new_bins[label] = v  # type:ignore
        #    self.summary[k]["bins"] = new_bins

        # put the profile into the context so it gets passed along
        context["mabel:profile"] = self.summary
        return context

    def __str__(self):
        def _inner():
            for field in self.fields:
                field_summary = self.summary[field]
                if field_summary["nulls"] == field_summary["items"]:
                    yield f"[---] {field:20} [count] {field_summary['items']} EMPTY"
                elif field_summary["type"] == "numeric":
                    yield f"[num] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} [range] {ProfileDataOperator.short_form(field_summary.get('min'))} to {ProfileDataOperator.short_form(field_summary.get('max'))} [mean] {ProfileDataOperator.short_form(field_summary['mean'])} >{draw_histogram_bins([v for k,v in field_summary['bins'].items()])}<"
                elif field_summary["type"] == "date":
                    yield f"[dte] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} [range] {field_summary.get('min')} to {field_summary.get('max')} >{draw_histogram_bins([v for k,v in field_summary['bins'].items()])}<"
                elif field_summary["type"] == "other":
                    yield f"[oth] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%}"
                elif field_summary["type"] == "string":
                    yield f"[str] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} [unique] {field_summary['unique_values']} [longest] {field_summary['max_length']} [shortest] {field_summary['min_length']}"
                elif field_summary["type"] == "enum":
                    yield f"[enm] {field:20} [count] {field_summary['items']} [empty] {(field_summary['nulls'] / field_summary['items']):.1%} {ProfileDataOperator.enum_summary(field_summary['values'])}"

        return "\n".join(_inner())

    def __repr__(self):
        return json.serialize(self.summary, indent=True)

    @staticmethod
    def redistribute_bins(bins, number_of_bins=100):
        """
        Redistribute binned data into fewer bins

        This results in items being put into incorrect bins, especially
        when the data is ordered.
        """
        mn = min([l for l, h in bins])
        mx = max([h for l, h in bins])

        bin_size = ((mx - mn) + 1) // number_of_bins

        new_bins = {}
        for counter in range(number_of_bins - 1):
            new_bins[
                (mn + (bin_size * counter), mn + (bin_size * (counter + 1)) - 1)
            ] = 0
        new_bins[(mn + (bin_size * (counter + 1)), mx)] = 0

        for old_bounds in bins:
            old_lower, old_upper = old_bounds
            old_mid = (old_lower + old_upper) // 2
            binned = False
            for new_bounds in new_bins:
                new_lower, new_upper = new_bounds
                if old_mid >= new_lower and old_mid <= new_upper:
                    new_bins[new_bounds] += bins[old_bounds]
                    binned = True
                    break
            if binned == False:
                print("miss", old_mid, mn, mx)

        return new_bins

    @staticmethod
    def get_type(validators):
        """
        Determine the data type so we can use the correct profiler
        """
        try:
            val = [str(v).split(".")[0].split(" ")[1] for v in validators]
            val = [v for v in val if not v in ["is_null", "is_other"]]
            val = val.pop()
        except (KeyError, ValueError, IndexError):
            val = "other"

        if val in ["is_numeric"]:
            return "numeric"
        if val in ["is_string", "is_cve"]:
            return "string"
        if val in ["is_valid_enum", "is_boolean"]:
            return "enum"
        if val in ["is_date"]:
            return "date"

        return "other"

    @staticmethod
    def date_from_epoch(seconds, form="%Y-%m-%d %H:%M:%S"):
        """
        Convert from milliseconds since epoch to a date string
        """
        import datetime

        return datetime.datetime.fromtimestamp(seconds).strftime(form)

    @staticmethod
    def profile_numeric_data(collector, value):
        """
        Profiler for numeric data, it collects:

        - minimim value
        - maximum value
        - cummulative sum
        - mean
        - approximate distribution

        This method take single value at a time to build the profile
        """
        if collector.get("max", value) <= value:
            collector["max"] = value
        if collector.get("min", value) >= value:
            collector["min"] = value
        collector["cumsum"] = collector.get("cumsum", 0) + value
        collector["mean"] = collector["cumsum"] / collector["items"]

        #        if collector.get("bins") is None:
        #            collector["bins"] = {}
        #        binned = False
        #        for bounds in collector["bins"]:
        #            bottom, top = bounds
        #            if value >= bottom and value <= top:
        #                collector["bins"][bounds] += 1
        #                binned = True
        #        if not binned:
        #            collector["bins"][(value, value)] = 1
        #
        #        if len(collector["bins"]) > 5007:  # first prime over 10000
        #            collector["bins"] = ProfileDataOperator.redistribute_bins(
        #                collector["bins"], 101
        #            )  # first prime over 100

        return collector

    @staticmethod
    def profile_string_data(collector, value):
        """
        Profiler for string data, it collects:

        - minimim length
        - maximum length
        - unique values (capped upper limit)

        This method take single value at a time to build the profile
        """
        if collector.get("max_length", len(value)) <= len(value):
            collector["max_length"] = len(value)
        if collector.get("min_length", len(value)) >= len(value):
            collector["min_length"] = len(value)
        if collector.get("bloom_filter") is None:
            collector["unique_values"] = []
            collector["bloom_filter"] = BloomFilter(
                number_of_elements=BLOOM_FILTER_SIZE, fp_rate=0.01
            )
            collector["approx_unique_values"] = 0

        # if we have only a few values, count them all
        if len(collector["unique_values"]) < ACCURATE_THRESHOLD:
            hashed = hash(value)
            if not hashed in collector["unique_values"]:
                collector["unique_values"].append(hashed)

        # when we have a lot of values, use a bloom filter
        if value not in collector["bloom_filter"]:
            collector["bloom_filter"].add(value)
            collector["approx_unique_values"] += 1

        return collector

    @staticmethod
    def short_form(num):
        if not isinstance(num, (int, float)):
            return num
        if isinstance(num, int) and abs(num) < 10000000:
            return str(num)
        return f"{num:.2f}"

    @staticmethod
    def enum_summary(dic):
        s = {
            k: v for k, v in sorted(dic.items(), key=lambda item: item[1], reverse=True)
        }
        cumsum = sum([v for k, v in dic.items()])
        eliminated = 0
        result = ""
        for index, item in enumerate(s):
            if index == 2:
                break
            result += f"`{item}`: {(s[item] / cumsum):.1%} "
            eliminated += s[item]
        if eliminated < cumsum:
            result += f"[other]: {(cumsum - eliminated) / cumsum:.1%}"
        return result



from typing import List, Callable


def select_record_fields(record: dict, fields: List[str]) -> dict:
    """
    Selects a subset of fields from a dictionary. If the field is not present
    in the dictionary it defaults to None.

    Parameters:
        record: dictionary
            The dictionary to select from
        fields: list of strings
            The list of the field names to select

    Returns:
        dictionary
    """
    return {k: record.get(k, None) for k in fields}


def order(record: dict) -> dict:
    """
    Sort a dictionary by its keys.

    Parameters:
        record: dictionary
            The dictionary to sort

    Returns:
        dictionary
    """
    return dict(sorted(record.items()))


def set_value(record: dict, field_name: str, setter: Callable) -> dict:
    """
    Sets the value of a column to either a fixed value or as the result of a
    function which recieves the row as a parameter.

    Parameters:
        record: dictionary
            The dictionary to update
        field_name: string
            The field to create or update
        setter: callable or any
            A function or constant to update the field with

    Returns:
        dictionary
    """
    copy = record.copy()
    if callable(setter):
        copy[field_name] = setter(copy)
    else:
        copy[field_name] = setter
    return copy



"""
Reader is about 300% faster (8 processes) than the default serial reader.

Problems
- The .limit method causes this reader to never end ()
"""
import multiprocessing
import os
import time
from ...formats import dictset
from ....logging import get_logger


MAXIMUM_SECONDS_PROCESSES_CAN_RUN = 600  # 5 minutes
TERMINATE_SIGNAL = -1


def _inner_process(
    flag, reader, source_queue, reply_queue, parser, where
):  # pragma: no cover

    try:
        source = source_queue.get(timeout=0.1)
    except Exception:  # pragma: no cover
        source = None

    while source is not None and flag.value != TERMINATE_SIGNAL:
        data = reader.get_records(source)
        data = parser(data)
        if where is not None:
            data = filter(where, data)
        for chunk in dictset.page_dictset(data, 256):
            # wait - to save memory we have limited number of slots
            reply_queue.put(chunk, timeout=60)
        try:
            source = source_queue.get(timeout=0.1)
        except Exception:  # pragma: no cover
            source = None

    flag.value = TERMINATE_SIGNAL
    get_logger().debug("Terminating background process")


def processed_reader(items_to_read, reader, parser, where):  # pragma: no cover

    if os.name == "nt":  # pragma: no cover
        raise NotImplementedError(
            "Reader Multi Processing not available on Windows platforms"
        )

    process_pool = []

    send_queue = multiprocessing.Queue()
    for item in items_to_read:
        send_queue.put(item)

    # limit the number of slots
    slots = min(8, len(items_to_read), multiprocessing.cpu_count())
    reply_queue = multiprocessing.Queue(slots * 16)

    for _ in range(slots):
        flag = multiprocessing.Value("i", 1 - TERMINATE_SIGNAL)
        process = multiprocessing.Process(
            target=_inner_process,
            args=(flag, reader, send_queue, reply_queue, parser, where),
        )
        process.daemon = True
        process.start()
        process_pool.append(flag)

    process_pool = set(process_pool)

    process_start_time = time.time()

    # if we're searching for a rare term, it will go seconds without returning
    # any results, but we shouldn't wait forever
    while any({t.value == 1 - TERMINATE_SIGNAL for t in process_pool}) or not (
        reply_queue.empty()
    ):
        try:
            records = reply_queue.get(timeout=1)
            yield from records
        except:  # nosec
            if time.time() - process_start_time > MAXIMUM_SECONDS_PROCESSES_CAN_RUN:
                get_logger().debug(
                    f"Sending TERMINATE to long running multi-processed processes after {MAXIMUM_SECONDS_PROCESSES_CAN_RUN} seconds total run time"
                )
                for flag in process_pool:
                    flag.value = TERMINATE_SIGNAL



"""
Trace Blocks

As data moves between the flows, Trace Blocks is used to create a record of
operation being run.

It uses an approach similar to a block-chain in that each block includes a
hash of the previous block.

The block contains a hash of the data, the name of the operation, a 
programatically determined version of the code that was run, a timestamp and
a hash of the last block.

This isn't distributed, but the intention is that the trace log writes the
block hash at the time the data is processed which this Class creating an
independant representation of the trace. In order to bypass this control,
the user must update the audit log and this trace block.

#nodoc - don't add to the documentation wiki
"""
import json
import hashlib
import datetime
from typing import Optional
from ...utils import entropy

serialize = json.dumps  # prevent circular imports

EMPTY_HASH = "0" * 64


class TraceBlocks:

    __slots__ = ("blocks", "proof")

    def __init__(
        self, uuid="00000000-0000-0000-0000-000000000000", proof: Optional[str] = None
    ):
        """
        Create block chain and seed with the UUID.

        Parameters:
            uuid: string (UUID) (optional but strongly recommended)
                Unique identifier for the run
            proof: string (optional)
                The set of valid final characters for proofing as a string.
                The default is None which does no proofing. If you set this
                wrong, you could end with an endless loop.
        """
        self.proof = proof
        self.blocks = []
        self.blocks.append(
            {"block": 1, "timestamp": datetime.datetime.now().isoformat(), "uuid": uuid}
        )

    def add_block(self, **kwargs):
        """
        Add a new block to the chain.
        """
        previous_block = self.blocks[-1]
        previous_block_hash = self.hash(previous_block)

        # proof is what makes mining for bitcoin so hard, we're setting a low
        # target of the last character being a 0 or a 5 (i.e. 1/5 chance)
        # if you wanted to make this harder, set a different rule to exit
        # while loop. Setting this proof to be harder will impact performance
        # as finding a value to satify the proof will block processing.

        if self.proof:
            proof = str(entropy.random_int())
            while (
                self.hash("".join([proof, previous_block_hash]))[-1] not in self.proof
            ):
                proof = str(entropy.random_int())

        block = {
            "block": len(self.blocks) + 1,
            "timestamp": datetime.datetime.now().isoformat(),
            "previous_block_hash": previous_block_hash,
            **kwargs,
        }
        if self.proof:
            block["proof"] = proof
        self.blocks.append(block)

    def __str__(self):
        return serialize(self.blocks, indent=True)

    def __repr__(self):
        return f"<TraceBlocks with {len(self.blocks)} blocks>"

    def hash(self, block):
        try:
            bytes_object = serialize(block)
        except:
            bytes_object = block
        raw_hash = hashlib.sha256(bytes_object.encode())
        hex_hash = raw_hash.hexdigest()
        return hex_hash



import datetime
from mabel.errors.invalid_combination_error import InvalidCombinationError
from pydantic import BaseModel  # type:ignore
from typing import Any, Optional, Union, List
from dateutil import parser
from .internals.blob_writer import BlobWriter
from ..validator import Schema  # type:ignore
from ...utils import paths
from ...errors import ValidationError, InvalidDataSetError
from ...logging import get_logger
from ...data.formats import json


class SimpleWriter:
    def _get_writer_date(self, date):
        # default to today if not given a date
        batch_date = datetime.datetime.now()
        if isinstance(date, datetime.date):
            batch_date = date  # type:ignore
        if isinstance(date, str):
            batch_date = parser.parse(date)
        return batch_date

    def __init__(
        self,
        *,
        schema: Optional[Union[Schema, list]] = None,
        set_of_expectations: Optional[list] = None,
        format: str = "zstd",
        date: Any = None,
        **kwargs,
    ):
        """
        Simple Writer provides a basic writer capability.
        """
        dataset = kwargs.get("dataset", "")
        if kwargs.get("to_path"):  # pragma: no cover
            get_logger().warning(
                "DEPRECATION: Writer 'to_path' parameter has been replaced with 'dataset' "
            )
            dataset = kwargs.pop("to_path")

        if "BACKOUT" in dataset:
            InvalidDataSetError(
                "BACKOUT is a reserved word and cannot be used in Dataset names"
            )
        if dataset.endswith("/"):
            InvalidDataSetError("Dataset names cannot end with /")
        if "{" in dataset or "}" in dataset:
            InvalidDataSetError("Dataset names cannot contain { or }")
        if "%" in dataset:
            InvalidDataSetError("Dataset names cannot contain %")

        self.schema = None
        if isinstance(schema, list):
            schema = Schema(schema)
        if isinstance(schema, Schema):
            self.schema = schema

        #        if self.schema:
        #            get_logger().warning("Schema/Fields will be replaced by Expectations in a future version")

        # self.expectations = None
        # if set_of_expectations:
        #    self.expectations = Expectations(set_of_expectations=set_of_expectations)

        # if self.schema is not None and self.expectations is not None:
        #    raise InvalidCombinationError(
        #        "Writer cannot have `schema`\`fields` and `expectations` set at the same time."
        #    )

        self.finalized = False
        self.batch_date = self._get_writer_date(date)

        self.dataset_template = dataset
        if "{date" not in self.dataset_template and not kwargs.get("raw_path", False):
            self.dataset_template += "/{datefolders}"
            self.raw_path = True

        self.dataset = paths.build_path(self.dataset_template, self.batch_date)

        # add the values to kwargs
        kwargs["raw_path"] = True  # we've just added the dates
        kwargs["format"] = format
        kwargs["dataset"] = self.dataset

        arg_dict = kwargs.copy()
        arg_dict["dataset"] = f"{self.dataset}"
        arg_dict[
            "inner_writer"
        ] = f"{arg_dict.get('inner_writer', {}).__name__}"  # type:ignore
        get_logger().debug(json.serialize(arg_dict))

        # default index
        kwargs["index_on"] = kwargs.get("index_on", [])

        # create the writer
        self.blob_writer = BlobWriter(**kwargs)

        self.records = 0

    def append(self, record: Union[dict, BaseModel]):
        """
        Append a new record to the Writer

        Parameters:
            record: dictionary or pydantic.BaseModel
                The record to append to the Writer

        Returns:
            integer
                The number of records in the current blob
        """
        if isinstance(record, BaseModel):
            record = record.dict()
        elif self.schema and not self.schema.validate(
            subject=record, raise_exception=False
        ):
            raise ValidationError(
                f"Schema Validation Failed ({self.schema.last_error})"
            )
        # elif self.expectations:
        #    self.expectations.test_record(record)

        self.blob_writer.append(record)
        self.records += 1

    def __del__(self):
        if hasattr(self, "finalized") and not self.finalized:
            get_logger().error(
                f"{type(self).__name__} has not been finalized - data may be lost, make sure you call .finalize()"
            )

    def finalize(self):
        self.finalized = True
        try:
            return self.blob_writer.commit()
        except Exception as e:
            get_logger().error(
                f"{type(self).__name__} failed to close pool: {type(e).__name__} - {e}"
            )
        return None



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import StreamWriter
from ...adapters.minio import MinIoWriter


class MinIoStreamWriterOperator(BaseWriterOperator):
    def __init__(
        self, *, end_point, access_key, secret_key, secure: bool = False, **kwargs
    ):
        """
        Streaming Write Operator for MinIo.

        Uses StreamWriter for logic and MinIoWriter for IO
        """
        kwargs["end_point"] = end_point
        kwargs["access_key"] = access_key
        kwargs["secret_key"] = secret_key
        kwargs["secure"] = secure

        super().__init__(writer=StreamWriter, inner_writer=MinIoWriter, **kwargs)



import re
import abc  # abstract base class library
import sys
import time
import inspect
import hashlib
import datetime
import functools
from typing import List
from ...logging import get_logger  # type:ignore
from ...errors import render_error_stack, TimeExceeded
from ...data.formats.json import parse, serialize


# inheriting ABC is part of ensuring that this class only ever
# interited from
class BaseOperator(abc.ABC):
    @staticmethod
    @functools.lru_cache(1)
    def sigterm():
        """
        Create a termination signal - SIGTERM, when the data record is this, we do any
        closure activities.

        Change this each cycle to reduce the chances of it being sent accidently (it
        shouldn't but what could happen is the payload is saved from one job and fed to
        another)
        """
        full_hash = hashlib.sha256(datetime.datetime.now().isoformat().encode())
        return "SIGTERM-" + full_hash.hexdigest()

    def __init__(self, *args, **kwargs):
        """
        All Operators should inherit from this class, it will help ensure a common
        structure to Operator classes and provide some common functionality and
        interfaces.

        This Base Operator is relatively complex but serves to simplify the
        implementation of functional Operators so engineers can focus on handling data,
        not getting a pipeline to work.

        You are expected to override the execute method with the logic of the Operator,
        this should return None, if the pipeline should terminate at this Operator (for
        example, if the Operator filters records), return a tuple of (data, context),
        or a generator/list of (data, context).

        Parameters:
            retry_count: integer (optional)
                The number of times to attempt a Operator before aborting, this
                defaults to 2 and is limited between 1 and 5
            retry_wait: integer (optional)
                The number of seconds to wait between retries, this defaults to 5 and
                is limited between 1 and 300
            rolling_failure_window: integer (optional)
                The number of previous Operators to remember the success/failure of,
                when >50% of the Operators in this window are failures, the job aborts.
                This defaults to 10 and is limited between 1 (single failure aborts)
                and 100
        """
        self.flow = None
        self.executions = 0  # number of times this Operator has been run
        self.execution_time_ns = 0  # nano seconds of cpu execution time
        self.errors = 0  # number of errors
        self.commencement_time = None  # the time processing started
        self.logger = get_logger()  # get the mabel logger

        self.name = self.__class__.__name__

        # read retry settings, clamp values to practical ranges
        self.retry_count = self._clamp(kwargs.get("retry_count", 2), 1, 5)
        self.retry_wait = self._clamp(kwargs.get("retry_wait", 5), 1, 300)
        rolling_failure_window = self._clamp(
            kwargs.get("rolling_failure_window", 10), 1, 100
        )
        self.last_few_results = [1] * rolling_failure_window  # track the last n results

        # Log the hashes of the __call__ and version methods
        call_hash = self._hash(inspect.getsource(self.__call__))[-8:]
        version_hash = self._hash(inspect.getsource(self.version))[-8:]
        self.logger.audit(
            {
                "operator": self.name,
                "call_hash": call_hash,
                "version_hash": self.version(),
                "version_method": version_hash,
            }
        )

    @abc.abstractmethod
    def execute(self, data: dict = {}, context: dict = {}):
        """
        YOU MUST OVERRIDE THIS METHOD

        This is where the main logic for the Operator is implemented.

        Parameters:
            data: Dictionary (or Any)
                The data to be processed, the Base Operator is opinionated for the data
                to be a dictionary, but any data type will work
            context: Dictionary
                Information to support the exeuction of the Operator

        Returns:
            None
                Do not continue further through the flow
            Tuple (data, context)
                The data to pass to the next Operator
            Iterable(data, context)
                Run the next Operator multiple times
        """
        pass  # pragma: no cover

    def finalize(self, context: dict = None):
        """
        Any code to finalize an operator should be implemented here.

        This should return the context to pass along
        """
        if not context:
            context = {}

        return context

    def __call__(self, data: dict = {}, context: dict = {}):
        """
        DO NOT OVERRIDE THIS METHOD

        This method wraps the `execute` method, which must be overridden, to add
        management of the execution such as sensors and retries.
        """
        if data == self.sigterm():
            outcome = None
            try:
                outcome = self.finalize(context)
            except Exception as err:
                self.logger.error(
                    f"Problem finalizing {self.name} - {type(err).__name__} - {err} - {context.get('run_id')}"
                )
            finally:
                if not outcome:
                    outcome = context
            return self.sigterm(), outcome

        if self.commencement_time is None:
            self.commencement_time = datetime.datetime.now()
        self.executions += 1
        attempts_to_go = self.retry_count
        while attempts_to_go > 0:
            try:
                start_time = time.perf_counter_ns()
                outcome = self.execute(data, context)
                my_execution_time = time.perf_counter_ns() - start_time
                self.execution_time_ns += my_execution_time
                # add a success to the last_few_results list
                self.last_few_results.append(1)
                self.last_few_results.pop(0)
                break
            except TimeExceeded as te:
                raise te
            except Exception as err:
                self.errors += 1
                attempts_to_go -= 1
                if attempts_to_go:
                    self.logger.error(
                        f"`{self.name}` - {type(err).__name__} - {err} - retry in {self.retry_wait} seconds ({context.get('run_id')})"
                    )
                    time.sleep(self.retry_wait)
                else:
                    error_log_reference = ""
                    error_reference = err
                    try:
                        error_payload = (
                            f"timestamp  : {datetime.datetime.today().isoformat()}\n"
                            f"operator   : {self.name}\n"
                            f"error type : {type(err).__name__}\n"
                            f"details    : {err}\n"
                            "========================================================================================================================\n"
                            f"{self._wrap_text(render_error_stack(), 120)}\n"
                            "=======================================================  context  ======================================================\n"
                            f"{self._wrap_text(str(context), 120)}\n"
                            "========================================================  data  ========================================================\n"
                            f"{self._wrap_text(str(data), 120)}\n"
                            "========================================================================================================================\n"
                        )
                        error_log_reference = self.error_writer(
                            error_payload
                        )  # type:ignore
                    except Exception as err:
                        self.logger.error(
                            f"Problem writing to the error bin, a record has been lost. `{self.name}`, {type(err).__name__} - {err} - {context.get('run_id')}"
                        )
                    finally:
                        # finally blocks are called following a try/except block regardless of the outcome
                        self.logger.alert(
                            f"`{self.name}` - {type(error_reference).__name__} - {error_reference} - tried {self.retry_count} times before aborting ({context.get('run_id')}) {error_log_reference}"
                        )
                        sys.exit(1)

                outcome = None
                # add a failure to the last_few_results list
                self.last_few_results.append(0)
                self.last_few_results.pop(0)

        if type(outcome).__name__ not in ("NoneType", "generator"):
            if not (
                isinstance(outcome, tuple)
                and len(outcome) == 2
                and isinstance(outcome[1], dict)
            ):
                raise ValueError(
                    f"Operator `{self.name}` returned type `{type(outcome).__name__}`, not the expected (value, context) values."
                )

        # message tracing
        if context.get("trace", False):
            data_hash = self._hash(data)
            context["execution_trace"].add_block(
                data_hash=data_hash,
                operator=self.name,
                operator_version=self.version(),
                execution_ns=my_execution_time,
                data_block=serialize(data),
            )
            self.logger.audit(f"{context.get('run_id')} {self.name} {data_hash}")

        # if there is a high failure rate, abort
        if sum(self.last_few_results) < (len(self.last_few_results) / 2):
            self.logger.alert(
                f"Failure Rate for `{self.name}` over last {len(self.last_few_results)} executions is over 50%, aborting."
            )
            sys.exit(1)

        return outcome

    def read_sensors(self):
        """
        Format data about the transformation, this can be overridden but it should
        include this information
        """
        response = {
            "operator": self.name,
            "version_hash": self.version(),
            "execution_count": self.executions,
            "error_count": self.errors,
            "execution_seconds": self.execution_time_ns / 1e9,
        }
        if self.executions == 0:
            self.logger.warning(f"`{self.name}` was never executed")
        if self.commencement_time:
            response["commencement_time"] = self.commencement_time.isoformat()
        return response

    @functools.lru_cache(1)
    def version(self):
        """
        DO NOT OVERRIDE THIS METHOD.

        The version of the Operator code, this is intended to facilitate
        reproducability and auditability of the pipeline. The version is the last 8
        characters of the hash of the source code of the 'execute' method. This
        removes the need for the developer to remember to increment a version variable.

        Hashing isn't security sensitive here, it's to identify changes rather than
        protect information.
        """
        source = inspect.getsource(self.execute)
        source = self._only_alpha_nums(source)
        full_hash = hashlib.sha256(source.encode())
        return full_hash.hexdigest()[-8:]

    def __del__(self):
        # do nothing - prevents errors if someone thinks they're being a good
        # citizen and calls super().__del__
        pass

    def error_writer(self, record):
        # this is a stub to be overridden
        raise ValueError("no error_writer attached")

    def __gt__(self, next_operators):
        """
        Alias for __rshift__, `>`
        """
        self.logger.warning("Use of `>` to build flows will be replaced with `>>`")
        return self.__rshift__(next_operators)

    def __rshift__(self, next_operators):
        from ...flows.flow import Flow

        if not isinstance(next_operators, (list, set)):
            next_operators = [next_operators]
        if self.flow:
            # if I have a graph already, build on it
            flow = self.flow
        else:
            # if I don't have a graph, create one
            flow = Flow()
            flow.add_operator(f"{self.name}-{id(self)}", self)

        for operator in next_operators:
            if isinstance(operator, Flow):
                # if we're pointing to a graph, merge with the current graph,
                # we need to find the node with no incoming nodes we identify
                # the entry-point
                flow.merge(operator)
                for entry_point in operator.get_entry_points():
                    flow.link_operators(
                        f"{self.name}-{id(self)}",
                        entry_point,
                    )
            elif issubclass(type(operator), BaseOperator):
                # otherwise add the node and edge and set the graph further down the
                # line
                flow.add_operator(f"{operator.name}-{id(operator)}", operator)
                flow.link_operators(
                    f"{self.name}-{id(self)}",
                    f"{operator.name}-{id(operator)}",
                )
                operator.flow = flow
            else:
                label = type(operator).__name__
                if hasattr(operator, "__name__"):
                    label = operator.__name__
                #  deepcode ignore return~not~implemented: appears to be false positive
                raise TypeError(
                    f"Operator {label} must inherit BaseOperator, this error also occurs when the Operator has not been correctly instantiated."
                )
        # this variable only exists to build the graph
        self.flow = None
        return flow

    def __rrshift__(self, previous_operators):
        from ...flows.flow import Flow

        if not isinstance(previous_operators, (list, set)):
            previous_operators = [previous_operators]
        if self.flow:
            # if I have a graph already, build on it
            flow = self.flow
        else:
            # if I don't have a graph, create one
            flow = Flow()
            flow.add_operator(f"{self.name}-{id(self)}", self)

        for operator in previous_operators:
            if isinstance(operator, Flow):
                # if we're pointing to a graph, merge with the current graph,
                # we need to find the node with no incoming nodes we identify
                # the entry-point
                flow.merge(operator)
                for exit in operator.get_exit_points():
                    flow.link_operators(
                        exit,
                        f"{self.name}-{id(self)}",
                    )
            elif issubclass(type(operator), BaseOperator):
                # otherwise add the node and edge and set the graph further down the
                # line
                flow.add_operator(f"{operator.name}-{id(operator)}", operator)
                flow.link_operators(
                    f"{operator.name}-{id(operator)}",
                    f"{self.name}-{id(self)}",
                )
                operator.flow = flow
            else:
                label = type(operator).__name__
                if hasattr(operator, "__name__"):
                    label = operator.__name__
                #  deepcode ignore return~not~implemented: appears to be false positive
                raise TypeError(
                    f"Operator {label} must inherit BaseOperator, this error also occurs when the Operator has not been correctly instantiated."
                )
        # this variable only exists to build the graph
        self.flow = None
        return flow

    def _clamp(self, value, low_bound, high_bound):
        """
        'clamping' is fixing a value within a range
        """
        value = min(value, high_bound)
        value = max(value, low_bound)
        return value

    def _only_alpha_nums(self, text):
        pattern = re.compile(r"[\W_]+")
        return pattern.sub("", text)

    def _hash(self, block):
        try:
            bytes_object = serialize(block)
        except:
            bytes_object = str(block)
        raw_hash = hashlib.sha256(bytes_object.encode())
        hex_hash = raw_hash.hexdigest()
        return hex_hash

    def _wrap_text(self, text, line_len):
        from textwrap import fill

        def _inner(text):
            for line in text.splitlines():
                yield fill(line, line_len)

        return "\n".join(list(_inner(text)))

    def __repr__(self):
        return f"<{self.name}, version: {self.version()}>"



import time
from ..flows.internals.base_operator import BaseOperator
from ..data import Reader
from ..data.formats import json
from ..errors.time_exceeded import TimeExceeded


class ReaderOperator(BaseOperator):
    def __init__(
        self,
        *args,
        time_out: float = 315569652.0,
        signal_format: str = "{cursor}",
        **kwargs
    ):
        """
        Read through a dataset and emit each row as a new record. Includes optional
        functionality to stop after a set amount of time.

        Parameters:
            time_out: float (optional)
                The number of seconds to run before bailing, default is 1 year
                (non-leap)
            signal_format: string (optional)
                The format of the detail in the exception. The default is '{cursor}'
                which is replaced with the reader cursor value at the time of the
                time_out.
            other values as per the inner_reader.
        Yields:
            tuple

        Raises:
            TimeExceeded
        """
        super().__init__(*args, **kwargs)
        self.time_out = time_out + time.time()
        self.signal_format = signal_format
        self.reader = Reader(**kwargs)

    def execute(self, data, context):
        for row in self.reader:
            if time.time() > self.time_out:
                signal = self.signal_format.replace(
                    "{cursor}", json.serialize(self.reader.cursor)
                )
                raise TimeExceeded(signal)
            yield row, context



from typing import Callable, Iterator
from mabel.data.formats import dictset as ds


class Groups:

    __slots__ = "_groups"

    def __init__(self, dictset: Iterator, column: str, dedupe: bool = False):
        """
        Group By functionality for Iterables of Dictionaries
        Parameters:
            dictset: Iterable of dictionaries
                The dataset to perform the Group By on
            column: string
                The name of the field to group by
            dedupe: bool
                Remove duplicate values from the groups
        Returns:
            Groups
        Warning:
            The 'Groups' object holds the entire dataset in memory so is unsuitable
            for large datasets.
        """
        groups: dict = {}
        if dedupe:
            dictset = ds.drop_duplicates(dictset)
        for item in dictset:
            my_item = item.copy()
            key = my_item.pop(column, None)
            if not key in groups:
                groups[key] = []
            groups[key].append(my_item)
        if dedupe:
            for group in groups:
                groups[group] = {
                    frozenset(item.items()): item for item in groups[group]
                }.values()
        self._groups = groups

    def count(self, group=None):
        """
        Count the number of items in groups
        Parameters:
            group: string (optional)
                If provided, return the count of just this group
        Returns:
            if a group is provided, return an integer
            if no group is provided, return a dictionary
        """
        if group is None:
            return {x: len(y) for x, y in self._groups.items()}
        else:
            try:
                return [len(y) for x, y in self._groups.items() if x == group].pop()
            except IndexError:
                return 0

    def aggregate(self, column, method):
        """
        Applies an aggregation function by group.
        Parameters:
            column: string
                The name of the field to aggregate on
            method: callable
                The function to aggregate with
        Returns:
            dictionary
        Examples:
            maxes = grouped.aggregate('age', max)
            means = grouped.aggregate('age', maths.mean)
        """
        response = {}
        for key, items in self._groups.items():
            values = [
                item.get(column) for item in items if item.get(column) is not None
            ]
            response[key] = method(values)
        return response

    def apply(self, method: Callable):
        """
        Apply a function to all groups
        Parameters:
            method: callable
                The function to apply to the groups
        Returns:
            dictionary
        """
        return {key: method(items) for key, items in self._groups.items()}

    def __len__(self):
        """
        Returns the number of groups in the set.
        """
        return len(self._groups)

    def __repr__(self):
        """
        Returns the group names
        """
        return f"Group of {len(self)} items"

    def __getitem__(self, item):
        """
        Selector access to groups, e.g. Groups["Group Name"]
        Note that Groups["Group 1", "Group 2"] creates a group with just those items
        """
        if isinstance(item, (tuple, list)):
            newg = Groups([], None)
            for entry in item:
                if entry in self._groups:
                    newg._groups[entry].append(self._groups[entry])
            return newg
        else:
            return SubGroup(self._groups.get(item))


class SubGroup:

    __slots__ = "values"

    def __init__(self, values):
        self.values = values or []

    def __getitem__(self, item):
        """
        Selector access to a value in a group, support arrays
        """
        if isinstance(item, tuple):
            return list(ds.select_from(self.values, columns=item))
        else:
            return ds.extract_column(self.values, item)

    def __len__(self):
        return len(self.values)

    def __repr__(self):
        return f"SubGroup of {len(self)} items"



from setuptools import setup, find_packages  # type:ignore

with open("mabel/version.py", "r") as v:
    vers = v.read()
exec(vers)  # nosec

with open("README.md", "r") as rm:
    long_description = rm.read()

setup(
    name="mabel",
    version=__version__,
    description="Python Data Libraries",
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Joocer",
    packages=find_packages(include=["mabel", "mabel.*"]),
    url="https://github.com/mabel-dev/mabel/",
    install_requires=["ujson", "python-dateutil", "zstandard", "mmh3", "pydantic"],
)



"""
Create .serialize and .parse methods to handle json operations

Where orjson is installed, the performance impact is nil, without orjson, parsing is
about as fast as ujson, however serialization is slower, although still faster than the
native json library. 
"""
from typing import Any, Union
import datetime

try:
    # if orjson is available, use it
    import orjson

    parse = orjson.loads

    def serialize(
        obj: Any, indent: bool = False, as_bytes: bool = False
    ) -> Union[str, bytes]:

        if as_bytes:
            if indent and isinstance(obj, dict):
                return orjson.dumps(
                    obj, option=orjson.OPT_INDENT_2 + orjson.OPT_SORT_KEYS
                )
            else:
                return orjson.dumps(obj, option=orjson.OPT_SORT_KEYS)

        # return a string
        if indent and isinstance(obj, dict):
            return orjson.dumps(
                obj, option=orjson.OPT_INDENT_2 + orjson.OPT_SORT_KEYS
            ).decode()
        else:
            return orjson.dumps(obj, option=orjson.OPT_SORT_KEYS).decode()


except ImportError:  # pragma: no cover
    # orjson doesn't install on 32bit systems so we need a backup plan
    # however, orjson and ujson have functional differences so we can't
    # just swap the references.
    import ujson

    def serialize(
        obj: Any, indent: bool = False, as_bytes: bool = False
    ) -> Union[str, bytes]:  # type:ignore
        def fix_fields(dt: Any) -> str:
            """
            orjson and ujson handles some fields differently,
            if one of those fields is detected, fix it.
            """
            if isinstance(dt, (datetime.date, datetime.datetime)):
                return dt.isoformat()
            return dt

        if isinstance(obj, dict):
            obj_copy = {k: fix_fields(v) for k, v in obj.items()}
        else:
            obj_copy = obj

        if as_bytes:
            if indent:
                return ujson.dumps(obj_copy, sort_keys=True, indent=2).encode()
            else:
                return ujson.dumps(obj_copy, sort_keys=True).encode()

        if indent:
            return ujson.dumps(obj_copy, sort_keys=True, indent=2)
        else:
            return ujson.dumps(obj_copy, sort_keys=True)

    parse = ujson.loads  # type:ignore



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import BatchWriter
from ...adapters.null import NullWriter


class NullBatchWriterOperator(BaseWriterOperator):
    def __init__(self, **kwargs):
        """ """
        super().__init__(writer=BatchWriter, inner_writer=NullWriter, **kwargs)



# type:ignore
import io
import glob
import datetime
from typing import Iterable, Tuple, Optional, List
from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...utils import paths, common
from ...logging import get_logger
from ...errors import InvalidReaderConfigError, MissingDependencyError

try:
    import paho.mqtt.client as mqtt

    mqtt_installed = True
except ImportError:  # pragma: no cover
    mqtt_installed = False


# pip install paho-mqtt


class MqttReader(BaseInnerReader):
    def __init__(self, **kwargs):

        if not mqtt_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`paho-mqtt` is missing, please install or include in requirements.txt"
            )

        super().__init__(**kwargs)
        self.dataset = kwargs.get("dataset")
        self.host = kwargs.get("host")
        if not self.host:
            raise InvalidReaderConfigError("MqttReader requires `host` to be set.")
        self.port = int(kwargs.get("port", 1883))

        self.messages = []

    def on_connect(self, client, userdata, flags, rc):
        get_logger().debug(
            f"Connected to mqtt server {self.host} with result code {rc}"
        )
        client.subscribe(self.dataset)

    def on_message(self, client, userdata, msg):
        self.messages.append(msg.payload)

    def get_list_of_blobs(self):
        """
        Override this method to just return the queue we're interested in
        """
        return [self.dataset]

    def get_blobs_at_path(self, path):
        # not used but must be present
        pass

    def get_blob_stream(self, blob_name: str) -> io.IOBase:
        # not used but must be present
        pass

    def get_records(
        self, blob_name: str, rows: Optional[Iterable[int]] = None
    ) -> Iterable[str]:
        """
        Override this method to just return records as they appear
        on the queue
        """
        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.connect(self.host, self.port, 60)

        while True:
            self.client.loop(timeout=1.0)
            if self.messages:
                yield from self.messages
            self.messages.clear()

    def __del__(self):
        self.client.loop_stop()



from ..flows.internals.base_operator import BaseOperator


def match_all(data):
    return True


class FilterOperator(BaseOperator):
    def __init__(self, condition=None):
        """
        Filters records, returns the record for matching records and returns 'None'
        for non-matching records.

        The Filter Operator takes one configuration item at creation - a _Callable_
        (function) which takes the data as a dictionary as it's only parameter and
        returns `True` to retain the record, or `False` to not pass the record
        through to the next Operator.

        Example:
            fo = FilterOperator(condition=lambda r: r.get('severity') == 'high')

        Note:
            The condition does not need to be lambda, it can be any _Callable_
            including methods.
        """
        super().__init__()
        if not condition:
            self.logger.warning(
                "FilterOperator missing expected named parameter `condition`."
            )
            condition = match_all
        self.condition = condition

    def execute(self, data={}, context={}):
        if self.condition(data):
            return data, context
        return None



import io
from ...data.writers.internals.base_inner_writer import BaseInnerWriter
from ...errors import MissingDependencyError

try:
    from minio import Minio  # type:ignore

    minio_installed = True
except ImportError:  # pragma: no cover
    minio_installed = False


class MinIoWriter(BaseInnerWriter):
    def __init__(
        self,
        *,
        end_point: str,
        access_key: str,
        secret_key: str,
        secure: bool = False,
        **kwargs
    ):

        if not minio_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`minio` is missing, please install or include in requirements.txt"
            )
        super().__init__(**kwargs)

        self.client = Minio(end_point, access_key, secret_key, secure=secure)
        self.filename = self.filename_without_bucket

    def commit(self, byte_data, override_blob_name=None):

        # if we've been given the filename, use that, otherwise get the
        # name from the path builder
        if override_blob_name:
            blob_name = override_blob_name
        else:
            blob_name = self._build_path()

        self.client.put_object(
            self.bucket, blob_name, io.BytesIO(byte_data), len(byte_data)
        )

        return blob_name



"""
A MongoDB Reader

This is a light-weight and prototype MongoDB reader.
"""
import io
from typing import Iterable, Optional, List
from ...data.readers.internals.base_inner_reader import BaseInnerReader
from ...errors import MissingDependencyError

try:
    import pymongo  # type:ignore

    mongo_installed = True
except ImportError:  # pragma: no cover
    mongo_installed = False


class MongoDbReader(BaseInnerReader):
    def __init__(self, connection_string: str, **kwargs):

        if not mongo_installed:  # pragma: no cover
            raise MissingDependencyError(
                "`pymongo` is missing, please install or include in requirements.txt"
            )

        super().__init__(**kwargs)

        connection = pymongo.MongoClient(connection_string)
        self.database = connection[kwargs.get("dataset")]

        # chunk size affects memory usage
        self.chunk_size: int = kwargs.get("chunk_size", 10)
        self.query: dict = kwargs.get("query", {})

    def get_list_of_blobs(self):
        return self.database.list_collection_names()

    def get_records(
        self, blob_name: str, rows: Optional[Iterable[int]] = None
    ) -> Iterable[str]:
        collection = self.database[blob_name]  # type:ignore
        chunks = self._iterate_by_chunks(
            collection, self.chunk_size, 0, query=self.query
        )
        for docs in chunks:
            yield from docs

    def _iterate_by_chunks(self, collection, chunksize=1, start_from=0, query={}):
        chunks = range(start_from, collection.find(query).count(), int(chunksize))
        num_chunks = len(chunks)
        for i in range(1, num_chunks + 1):
            if i < num_chunks:
                yield collection.find(query)[chunks[i - 1] : chunks[i]]
            else:
                yield collection.find(query)[chunks[i - 1] : chunks.stop]

    def get_blobs_at_path(self, path):
        # not used but must be present
        pass

    def get_blob_stream(self, blob_name: str) -> io.IOBase:
        # not used but must be present
        pass



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import StreamWriter
from ...adapters.google import GoogleCloudStorageWriter


class GoogleStorageStreamWriterOperator(BaseWriterOperator):
    def __init__(self, *, project: str = None, **kwargs):

        kwargs["project"] = project
        super().__init__(
            writer=StreamWriter, inner_writer=GoogleCloudStorageWriter, **kwargs
        )



from ...flows.internals.base_writer_operator import BaseWriterOperator
from ...data import BatchWriter
from ...adapters.google import GoogleCloudStorageWriter


class GoogleStorageBatchWriterOperator(BaseWriterOperator):
    def __init__(self, *, project: str = None, **kwargs):

        kwargs["project"] = project
        super().__init__(
            writer=BatchWriter, inner_writer=GoogleCloudStorageWriter, **kwargs
        )
